# -*- coding: utf-8 -*-
"""
Author: Xiang Yangcheng [https://github.com/Xayah-Hina]
GitHub: https://github.com/IME-lab-Hokudai/EPINF-NeuFlow
Last Modification Date: 2025-07-10
Description: The pytorch implementation of the paper "Efficient Physics Informed Neural Reconstruction of Diverse Fluid Dynamics from Sparse Observations".
License: Mozilla Public License 2.0 (MPL-2.0)
Copyright (c) 2025 Xiang Yangcheng, IME Lab, Hokkaido University, Japan.
"""
import torch


def advect_SL(q_grid, vel_world_prev, coord_3d_sim, dt, RK=1):
    """Advect a scalar quantity using a given velocity field.
    Args:
        q_grid: [X', Y', Z', C]
        vel_world_prev: [X, Y, Z, 3]
        coord_3d_sim: [X, Y, Z, 3]
        dt: float
        RK: int, number of Runge-Kutta steps
        y_start: where to start at y-axis
        proj_y: simulation domain resolution at y-axis
        use_project: whether to use Poisson solver
        project_solver: Poisson solver
        bbox_model: bounding box model
    Returns:
        advected_quantity: [X, Y, Z, 1]
        vel_world: [X, Y, Z, 3]
    """
    if RK == 1:
        vel_world = vel_world_prev.clone()
        vel_sim = vel_world  # [X, Y, Z, 3]
    elif RK == 2:
        vel_world = vel_world_prev.clone()  # [X, Y, Z, 3]
        # breakpoint()
        vel_sim = vel_world  # [X, Y, Z, 3]
        coord_3d_sim_midpoint = coord_3d_sim - 0.5 * dt * vel_sim  # midpoint
        midpoint_sampled = coord_3d_sim_midpoint * 2 - 1  # [X, Y, Z, 3]
        vel_sim = torch.nn.functional.grid_sample(vel_sim.permute(3, 2, 1, 0)[None], midpoint_sampled.permute(2, 1, 0, 3)[None], align_corners=True, padding_mode='zeros').squeeze(0).permute(3, 2, 1, 0)  # [X, Y, Z, 3]
    else:
        raise NotImplementedError
    backtrace_coord = coord_3d_sim - dt * vel_sim  # [X, Y, Z, 3]
    backtrace_coord_sampled = backtrace_coord * 2 - 1  # ranging [-1, 1]
    q_grid = q_grid[None, ...].permute([0, 4, 3, 2, 1])  # [N, C, Z, Y, X] i.e., [N, C, D, H, W]
    q_backtraced = torch.nn.functional.grid_sample(q_grid, backtrace_coord_sampled.permute(2, 1, 0, 3)[None, ...], align_corners=True, padding_mode='zeros')  # [N, C, D, H, W]
    q_backtraced = q_backtraced.squeeze(0).permute([3, 2, 1, 0])  # [X, Y, Z, C]
    return q_backtraced


def advect_maccormack(q_grid, vel_sim_prev, coord_3d_sim, dt):
    """
    Args:
        q_grid: [X', Y', Z', C]
        vel_world_prev: [X, Y, Z, 3]
        coord_3d_sim: [X, Y, Z, 3]
        dt: float
    Returns:
        advected_quantity: [X, Y, Z, C]
        vel_world: [X, Y, Z, 3]
    """
    q_grid_next = advect_SL(q_grid, vel_sim_prev, coord_3d_sim, dt)
    q_grid_back = advect_SL(q_grid_next, vel_sim_prev, coord_3d_sim, -dt)
    q_advected = q_grid_next + (q_grid - q_grid_back) / 2
    C = q_advected.shape[-1]
    for i in range(C):
        q_max, q_min = q_grid[..., i].max(), q_grid[..., i].min()
        q_advected[..., i] = q_advected[..., i].clamp_(q_min, q_max)
    return q_advected


def advect_density(den: torch.Tensor, vel: torch.Tensor, coord_3d_sim: torch.Tensor, dt: float, gt: torch.Tensor, gt_mask: torch.Tensor):
    den = advect_maccormack(den, vel, coord_3d_sim, dt)
    den[gt_mask] = gt[gt_mask]
    return den



# -*- coding: utf-8 -*-
"""
Author: Xiang Yangcheng [https://github.com/Xayah-Hina]
GitHub: https://github.com/IME-lab-Hokudai/EPINF-NeuFlow
Last Modification Date: 2025-07-10
Description: The pytorch implementation of the paper "Efficient Physics Informed Neural Reconstruction of Diverse Fluid Dynamics from Sparse Observations".
License: Mozilla Public License 2.0 (MPL-2.0)
Copyright (c) 2025 Xiang Yangcheng, IME Lab, Hokkaido University, Japan.
"""
import torch


class Plugin(torch.nn.Module):
    def __init__(self, name: str):
        super().__init__()
        self._name = name

    @property
    def name(self):
        return self._name

    def on_iter_start(self):
        pass

    def to(self, device):
        pass


class PluginLoss(Plugin):
    def __init__(self, name: str, weight):
        super().__init__(name)
        self._loss = torch.tensor(0.0)
        self._weight = weight
        self._enable = True

    def enable(self, enable: bool = True):
        self._enable = enable

    @property
    def loss(self):
        return self._loss

    @property
    def weight(self):
        return self._weight

    def on_iter_start(self):
        self._loss = torch.tensor(0.0, device=self._loss.device)

    def to(self, device):
        super().to(device)
        self._loss = self._loss.to(device=device)



# -*- coding: utf-8 -*-
"""
Author: Xiang Yangcheng [https://github.com/Xayah-Hina]
GitHub: https://github.com/IME-lab-Hokudai/EPINF-NeuFlow
Last Modification Date: 2025-07-10
Description: The pytorch implementation of the paper "Efficient Physics Informed Neural Reconstruction of Diverse Fluid Dynamics from Sparse Observations".
License: Mozilla Public License 2.0 (MPL-2.0)
Copyright (c) 2025 Xiang Yangcheng, IME Lab, Hokkaido University, Japan.
"""
import dataclasses
import os
import random
import typing

import rich.progress
import torch
import torchvision
import yaml

from .api import PIDataset, PIDatasetAPI
from .image_mask import process_video_full_mask


class PINeuFlow(PIDataset):
    """
    PRIVATE CLASS
    - This class and its subclasses are only used inside datasets package.
    """

    def __init__(self, dataset_path: str, dataset_type: typing.Literal['train', 'val', 'test'], downscale: int, use_fp16: bool, use_image_mask: bool):
        super().__init__()
        with open(os.path.join(dataset_path, 'scene_info.yaml'), 'r') as f:
            scene_info = yaml.safe_load(f)
            indices = torch.tensor(scene_info[dataset_type + '_indices'])

            # ---------- Load Base Information ----------
            poses = torch.stack([
                torch.tensor(cam["cam_transform"])
                for cam in scene_info["cameras"]
            ])[indices]
            focals = torch.tensor([
                float(cam["focal"]) * float(cam["width"]) / float(cam["aperture"])
                for cam in scene_info["cameras"]
            ])[indices]
            widths = torch.tensor([
                int(cam["width"])
                for cam in scene_info["cameras"]
            ])[indices]
            heights = torch.tensor([
                int(cam["height"])
                for cam in scene_info["cameras"]
            ])[indices]
            # ---------- Load Base Information ----------

            # ---------- Load videos ----------
            images = []
            with rich.progress.Progress() as progress:
                task = progress.add_task(f"[cyan]Loading Videos ({dataset_type})...", total=len(indices))
                for i in indices:
                    video_path = scene_info['videos'][i]
                    full_path = os.path.normpath(os.path.join(dataset_path, video_path))
                    video_tensor = torchvision.io.read_video(full_path, pts_unit='sec')[0].to(torch.float32) / 255.0
                    images.append(video_tensor)
                    progress.update(task, advance=1)
            images = torch.stack(images)
            # ---------- Load videos ----------

            if use_image_mask:
                # ---------- Load videos highlighted ----------
                images_highlighted = []
                with rich.progress.Progress() as progress:
                    task = progress.add_task(f"[cyan]Loading Highlighted Videos ({dataset_type})...", total=len(indices))
                    for i in indices:
                        video_path = scene_info['videos'][i]
                        highlighted_path = os.path.normpath(os.path.join(dataset_path, os.path.splitext(video_path)[0] + '_highlighted' + os.path.splitext(video_path)[1]))
                        if not os.path.exists(highlighted_path):
                            process_video_full_mask(
                                input_path=os.path.normpath(os.path.join(dataset_path, video_path)),
                                background_color=tuple([x * 255.0 for x in scene_info['background_color']]),
                                color_thresh=scene_info['color_thresh'],
                                dilation_pixels=scene_info['dilation_pixels'],
                            )
                        video_tensor = torchvision.io.read_video(highlighted_path, pts_unit='sec')[0].to(torch.float32) / 255.0
                        images_highlighted.append(video_tensor)
                        progress.update(task, advance=1)
                images_highlighted = torch.stack(images_highlighted)
                # ---------- Load videos highlighted ----------

                # ---------- Load videos masks ----------
                images_masks = []
                with rich.progress.Progress() as progress:
                    task = progress.add_task(f"[cyan]Loading Mask Videos ({dataset_type})...", total=len(indices))
                    for i in indices:
                        video_path = scene_info['videos'][i]
                        mask_path = os.path.normpath(os.path.join(dataset_path, os.path.splitext(video_path)[0] + '_mask' + os.path.splitext(video_path)[1]))
                        if not os.path.exists(mask_path):
                            process_video_full_mask(
                                input_path=os.path.normpath(os.path.join(dataset_path, video_path)),
                                background_color=tuple([x * 255.0 for x in scene_info['background_color']]),
                                color_thresh=scene_info['color_thresh'],
                                dilation_pixels=scene_info['dilation_pixels'],
                            )
                        video_tensor = torchvision.io.read_video(mask_path, pts_unit='sec')[0].to(torch.float32) / 255.0
                        images_masks.append(video_tensor)
                        progress.update(task, advance=1)
                images_masks = torch.stack(images_masks)
                # ---------- Load videos masks ----------

            # ---------- Downscale videos ----------
            V, T, H, W, C = images.shape
            if downscale != 1:
                images = images.permute(0, 1, 4, 2, 3).reshape(V * T, C, H, W)  # [V * T, C, H, W]
                if use_image_mask:
                    images_highlighted = images_highlighted.permute(0, 1, 4, 2, 3).reshape(V * T, C, H, W)  # [V * T, C, H, W]
                    images_masks = images_masks.permute(0, 1, 4, 2, 3).reshape(V * T, C, H, W)  # [V * T, C, H, W]
                H, W = H // downscale, W // downscale
                images = torch.nn.functional.interpolate(images, size=(H, W), mode='bilinear', align_corners=False).reshape(V, T, C, H, W).permute(0, 1, 3, 4, 2)  # [V, T, H, W, C]
                if use_image_mask:
                    images_highlighted = torch.nn.functional.interpolate(images_highlighted, size=(H, W), mode='bilinear', align_corners=False).reshape(V, T, C, H, W).permute(0, 1, 3, 4, 2)  # [V, T, H, W, C]
                    images_masks = torch.nn.functional.interpolate(images_masks, size=(H, W), mode='bilinear', align_corners=False).reshape(V, T, C, H, W).permute(0, 1, 3, 4, 2)  # [V, T, H, W, C]
                focals = focals / downscale
            # ---------- Downscale videos ----------

            # ---------- Load videos masks ----------
            images = images.permute(1, 0, 2, 3, 4)  # [T, V, H, W, C]
            if use_image_mask:
                images_highlighted = images_highlighted.permute(1, 0, 2, 3, 4)  # [T, V, H, W, C]
                images_masks = images_masks.permute(1, 0, 2, 3, 4)  # [T, V, H, W, C]
                images_gray = (
                        0.299 * images_masks[..., 0] +
                        0.587 * images_masks[..., 1] +
                        0.114 * images_masks[..., 2]
                )
                self._images_highlighted = images_highlighted
                self._images_masks = images_gray > 0.8
            else:
                self._images_highlighted = images
                self._images_masks = torch.ones_like(images[..., 0], dtype=torch.bool)
            # ---------- Load videos masks ----------

            # ---------- Prepare Data ----------
            self._images = images.to(dtype=torch.float16 if use_fp16 else torch.float32)
            self._poses = poses.to(dtype=torch.float16 if use_fp16 else torch.float32)
            self._focal = focals.to(dtype=torch.float16 if use_fp16 else torch.float32)
            self._width = widths.to(dtype=torch.float16 if use_fp16 else torch.float32)
            self._height = heights.to(dtype=torch.float16 if use_fp16 else torch.float32)

            self._near = float(scene_info['train_space']['near_std'])
            self._far = float(scene_info['train_space']['far_std'])

            self.times = torch.linspace(0, 1, steps=self.num_frames).to(dtype=torch.float16 if use_fp16 else torch.float32)

        self._use_image_mask = use_image_mask

    @property
    def images(self):
        return self._images

    @property
    def images_highlighted(self):
        return self._images_highlighted

    @property
    def images_masks(self):
        return self._images_masks

    @property
    def poses(self):
        return self._poses

    @property
    def focal(self):
        return self._focal

    @property
    def width(self):
        return self.images.shape[-2]

    @property
    def height(self):
        return self.images.shape[-3]

    @property
    def near(self):
        return self._near

    @property
    def far(self):
        return self._far

    @property
    def num_frames(self):
        return self.images.shape[0]

    @property
    def num_videos(self):
        return self.images.shape[1]

    @property
    def num_poses(self):
        return self.poses.shape[0]

    @property
    def num_times(self):
        return self.num_frames

    def transform(self, translate, scale):
        self._poses[:, :3, 3] = (self._poses[:, :3, 3] + torch.as_tensor(translate, dtype=self._poses.dtype, device=self._poses.device).view(1, 3)) * scale
        self._near = self._near * scale
        self._far = self._far * scale

    def __getitem__(self, index):
        time_shift = 0
        pose_index = random.randint(0, self.num_poses - 1)

        if index == 0 and time_shift <= 0:
            target_image = self.images[index, pose_index]
            target_time = self.times[index]
        elif index == self.images.shape[0] - 1 and time_shift >= 0:
            target_image = self.images[index, pose_index]
            target_time = self.times[index]
        else:
            if time_shift >= 0:
                target_image = (1 - time_shift) * self.images[index, pose_index] + time_shift * self.images[index + 1, pose_index]
                target_time = (1 - time_shift) * self.times[index] + time_shift * self.times[index + 1]
            else:
                target_image = (1 + time_shift) * self.images[index, pose_index] + (-time_shift) * self.images[index - 1, pose_index]
                target_time = (1 + time_shift) * self.times[index] + (-time_shift) * self.times[index - 1]

        return {
            'image': target_image,
            'image_highlighted': self.images_highlighted[index, pose_index],
            'image_mask': self.images_masks[index, pose_index],
            'pose': self.poses[pose_index],
            'focal': self.focal[pose_index],
            'time': target_time,
            'pose_index': pose_index,
            'frame_index': index,
        }

    def __len__(self):
        return len(self.images)


class PINeuFlowDataset(PIDatasetAPI):
    """
    BATCH DATA FORMAT:
        images: (N, H, W, C)
        images_masks: (N, H, W)
        poses: (N, 4, 4)
        focals: (N)
        times: (N)
        pose_indices: (N)
        frame_indices: (N)
    PUBLIC API:
        dataset_path: str
        aabb_std: list[min_x, min_y, min_z, max_x, max_y, max_z]
        bound_std: list[min_x, min_y, min_z, max_x, max_y, max_z]
        background_color: list[r, g, b]
        complex_background: bool
        num_train_poses: int
    """

    @dataclasses.dataclass
    class Config:
        base_data_dir: str = dataclasses.field(default='data/pi_neuflow/', metadata={'help': 'base data directory'})
        dataset: typing.Literal['scalar', 'torch', 'torch_avo', 'torch2', 'torch2_avo', 'fireplace', 'sphere', 'game'] = dataclasses.field(default='scalar', metadata={'help': 'dataset name'})
        downscale: int = dataclasses.field(default=1, metadata={'help': 'downscale factor for the dataset images'})
        downscale_val: int = dataclasses.field(default=2, metadata={'help': 'downscale factor for the validation images'})
        fp16: bool = dataclasses.field(default=True, metadata={'help': 'use mixed precision training'})

    def __init__(self, cfg: Config):
        super().__init__(cfg=cfg)

        self._background_color = None
        self._complex_background = None

        self._pre_translate = None
        self._pre_scale = None
        self._aabb_std = None
        self._bound_std = None

    def prepare_data(self):
        from huggingface_hub import snapshot_download
        snapshot_download(
            repo_id="XayahHina/PI-NeuFlow",
            repo_type="dataset",
            local_dir=self.cfg.base_data_dir,
        )

    def setup(self, stage: str):
        with open(os.path.join(self.cfg.base_data_dir, self.cfg.dataset, 'scene_info.yaml'), 'r') as f:
            scene_info = yaml.safe_load(f)
            self._background_color = scene_info['background_color']
            self._complex_background = int(scene_info['complex_background']) > 0

            train_space = scene_info['train_space']
            self._pre_translate = train_space['pre_translate']
            self._pre_scale = train_space['pre_scale']
            aabb_std_min = train_space['aabb_std_min']
            aabb_std_max = train_space['aabb_std_max']
            self._aabb_std = aabb_std_min + aabb_std_max
            self._bound_std = [-1.0, -1.0, -1.0, 1.0, 1.0, 1.0]
        super().setup(stage)

    def dataset(self, dataset_type: typing.Literal['train', 'val', 'test']) -> PIDataset:
        downscale = self.cfg.downscale_val if dataset_type == 'val' else self.cfg.downscale
        dataset = PINeuFlow(dataset_path=self.cfg.base_data_dir + self.cfg.dataset, dataset_type=dataset_type, downscale=downscale, use_fp16=self.cfg.fp16, use_image_mask=not self.complex_background)
        dataset.transform(translate=self.pre_translate, scale=self.pre_scale)
        return dataset

    @staticmethod
    def collate(batch: list):
        images = torch.stack([single['image'] for single in batch])
        images_highlighted = torch.stack([single['image_highlighted'] for single in batch])
        images_masks = torch.stack([single['image_mask'] for single in batch])
        poses = torch.stack([single['pose'] for single in batch])
        focals = torch.stack([single['focal'] for single in batch])
        times = torch.stack([single['time'] for single in batch])
        pose_indices = torch.tensor([single['pose_index'] for single in batch])
        frame_indices = torch.tensor([single['frame_index'] for single in batch])

        return {
            'images': images,
            'images_highlighted': images_highlighted,
            'images_masks': images_masks,
            'poses': poses,
            'focals': focals,
            'times': times,
            'pose_indices': pose_indices,
            'frame_indices': frame_indices,
        }

    @property
    def pre_translate(self):
        return self._pre_translate

    @property
    def pre_scale(self):
        return self._pre_scale

    @property
    def aabb_std(self):
        return self._aabb_std

    @property
    def bound_std(self):
        return self._bound_std

    @property
    def background_color(self):
        return self._background_color

    @property
    def complex_background(self):
        return self._complex_background



# -*- coding: utf-8 -*-
"""
Author: Xiang Yangcheng [https://github.com/Xayah-Hina]
GitHub: https://github.com/IME-lab-Hokudai/EPINF-NeuFlow
Last Modification Date: 2025-07-10
Description: The pytorch implementation of the paper "Efficient Physics Informed Neural Reconstruction of Diverse Fluid Dynamics from Sparse Observations".
License: Mozilla Public License 2.0 (MPL-2.0)
Copyright (c) 2025 Xiang Yangcheng, IME Lab, Hokkaido University, Japan.
"""
import dataclasses

import torch

from .api import RendererAPI
from ..models import NGPModel
from ..plugins import ImageMaskLoss


def sample_pdf(bins, weights, n_samples, det=False):
    # This implementation is from NeRF
    # bins: [B, T], old_z_vals
    # weights: [B, T - 1], bin weights.
    # return: [B, n_samples], new_z_vals

    # Get pdf
    weights = weights + 1e-5  # prevent nans
    pdf = weights / torch.sum(weights, -1, keepdim=True)
    cdf = torch.cumsum(pdf, -1)
    cdf = torch.cat([torch.zeros_like(cdf[..., :1]), cdf], -1)
    # Take uniform samples
    if det:
        u = torch.linspace(0. + 0.5 / n_samples, 1. - 0.5 / n_samples, steps=n_samples).to(weights.dtype).to(weights.device)
        u = u.expand(list(cdf.shape[:-1]) + [n_samples])
    else:
        u = torch.rand(list(cdf.shape[:-1]) + [n_samples]).to(weights.dtype).to(weights.device)

    # Invert CDF
    u = u.contiguous()
    inds = torch.searchsorted(cdf, u, right=True)
    below = torch.max(torch.zeros_like(inds - 1), inds - 1)
    above = torch.min((cdf.shape[-1] - 1) * torch.ones_like(inds), inds)
    inds_g = torch.stack([below, above], -1)  # (B, n_samples, 2)

    matched_shape = [inds_g.shape[0], inds_g.shape[1], cdf.shape[-1]]
    cdf_g = torch.gather(cdf.unsqueeze(1).expand(matched_shape), 2, inds_g)
    bins_g = torch.gather(bins.unsqueeze(1).expand(matched_shape), 2, inds_g)

    denom = (cdf_g[..., 1] - cdf_g[..., 0])
    denom = torch.where(denom < 1e-5, torch.ones_like(denom), denom)
    t = (u - cdf_g[..., 0]) / denom
    samples = bins_g[..., 0] + t * (bins_g[..., 1] - bins_g[..., 0])

    return samples


class NGPRenderer(RendererAPI):
    @dataclasses.dataclass
    class Config:
        num_sample_rays: int = dataclasses.field(default=1024 * 4, metadata={'help': 'number of rays to sample per batch'})
        num_depth_samples: int = dataclasses.field(default=1024, metadata={'help': 'number of depth samples per ray'})
        num_importance_samples: int = dataclasses.field(default=0, metadata={'help': 'number of importance samples per ray'})
        chunk: int = dataclasses.field(default=1024 * 4, metadata={'help': 'chunk size for rendering'})

        normalized_directions: bool = dataclasses.field(default=True, metadata={'help': 'whether to use normalized directions for rendering'})
        importance_samples: bool = dataclasses.field(default=True, metadata={'help': 'whether to use importance sampling for rendering'})

    def __init__(self, cfg: Config):
        super().__init__(cfg=cfg)
        self.register_plugin(plugin=ImageMaskLoss())

    @property
    def num_sample_rays(self):
        return self.cfg.num_sample_rays

    @property
    def normalized_directions(self):
        return self.cfg.normalized_directions

    @property
    def chunk(self):
        return self.cfg.chunk

    def render_impl(self,
                    model: NGPModel,
                    rays_o: torch.Tensor,
                    rays_d: torch.Tensor,
                    nears: torch.Tensor,
                    fars: torch.Tensor,
                    pose_indices: torch.Tensor,
                    ):
        num_sample_rays = rays_d.shape[0]

        xyz, z_vals = self.sample_points(
            rays_o=rays_o,
            rays_d=rays_d,
            nears=nears,
            fars=fars,
            num_depth_samples=self.cfg.num_depth_samples,
            use_perturb=model.training,
        )
        xyz_masked, mask = self.normalize_and_filter_xyz(xyz=xyz)

        if xyz_masked.numel() > 0:
            sigma_masked, geo_feat_masked = model.sigma(xyz_masked)
            sigma = self.assemble_masked_tensor(mask, sigma_masked, (num_sample_rays, self.cfg.num_depth_samples, 1))  # shape: (num_sample_rays, num_depth_samples, 1)

            if self.cfg.num_importance_samples > 0 and self.training:
                geo_feat = self.assemble_masked_tensor(mask, geo_feat_masked, (num_sample_rays, self.cfg.num_depth_samples, model.cfg.geo_feat_dim))  # shape: (num_sample_rays, num_depth_samples, geo_feat_dim)
                with torch.no_grad():
                    alpha = self.sigma_to_alpha(sigma=sigma, z_vals=z_vals)  # shape: (num_sample_rays, num_depth_samples)
                    weights = self.alpha_to_weights(alpha=alpha)  # shape: (num_sample_rays, num_depth_samples)
                    new_z_vals = sample_pdf(0.5 * (z_vals.squeeze(-1)[..., 1:] + z_vals.squeeze(-1)[..., :-1]), weights[:, 1:-1], self.cfg.num_importance_samples, det=not self.training).detach()  # shape: (num_sample_rays, num_importance_samples)
                    new_xyz = rays_o.unsqueeze(-2) + rays_d.unsqueeze(-2) * new_z_vals.unsqueeze(-1)  # shape: (num_sample_rays, num_importance_samples, 3)
                    new_xyz_masked, new_mask = self.normalize_and_filter_xyz(xyz=new_xyz)

                # override z_vals, and prepare z_index
                z_vals = torch.cat([z_vals.squeeze(-1), new_z_vals], dim=1)  # shape: (num_sample_rays, num_depth_samples + num_importance_samples,)
                z_vals, z_index = torch.sort(z_vals, dim=1)  # shape: (num_sample_rays, num_depth_samples + num_importance_samples,)
                z_vals = z_vals.unsqueeze(-1)  # shape: (num_sample_rays, num_depth_samples + num_importance_samples, 1)

                # override xyz
                xyz = torch.cat([xyz, new_xyz], dim=1)  # shape: (num_sample_rays, num_depth_samples + num_importance_samples, 3)
                xyz = torch.gather(xyz, dim=1, index=z_index.unsqueeze(-1).expand_as(xyz))  # shape: (num_sample_rays, num_depth_samples + num_importance_samples, 3)

                # override mask
                mask = torch.cat([mask.reshape(num_sample_rays, self.cfg.num_depth_samples), new_mask.reshape(num_sample_rays, self.cfg.num_importance_samples)], dim=1)
                mask = torch.gather(mask, dim=1, index=z_index)  # shape: (num_sample_rays, num_depth_samples + num_importance_samples)
                mask = mask.flatten()

                # override sigma_masked and geo_feat_masked
                new_sigma_masked, new_geo_feat_masked = model.sigma(new_xyz_masked)  # shape: (num_sample_rays * num_importance_samples, 1), (num_sample_rays * num_importance_samples, geo_feat_dim)
                new_sigma = self.assemble_masked_tensor(new_mask, new_sigma_masked, (num_sample_rays, self.cfg.num_importance_samples, 1))  # shape: (num_sample_rays, num_importance_samples, 1)
                tmp_sigma = torch.cat([sigma, new_sigma], dim=1)  # shape: (num_sample_rays, num_depth_samples + num_importance_samples, 1)
                sigma = torch.gather(tmp_sigma, dim=1, index=z_index.unsqueeze(-1).expand_as(tmp_sigma))  # shape: (num_sample_rays, num_depth_samples + num_importance_samples, 1)

                new_geo_feat = self.assemble_masked_tensor(new_mask, new_geo_feat_masked, (num_sample_rays, self.cfg.num_importance_samples, model.cfg.geo_feat_dim))  # shape: (num_sample_rays, num_importance_samples, geo_feat_dim)
                tmp_geo_feat = torch.cat([geo_feat, new_geo_feat], dim=1)  # shape: (num_sample_rays, num_depth_samples + num_importance_samples, geo_feat_dim)
                geo_feat = torch.gather(tmp_geo_feat, dim=1, index=z_index.unsqueeze(-1).expand_as(tmp_geo_feat))  # shape: (num_sample_rays, num_depth_samples + num_importance_samples, geo_feat_dim)
                geo_feat_masked = geo_feat.reshape(-1, model.cfg.geo_feat_dim)[mask]  # shape: (num_valid_points + new_num_valid_points, geo_feat_dim)

            view_dirs_masked = self.normalize_and_filter_view_dirs(view_dirs=rays_d, expand_shape=xyz.shape, mask=mask)
            rgb_masked = model.rgb(None, view_dirs_masked, geo_feat_masked)
            rgb = self.assemble_masked_tensor(mask, rgb_masked, xyz.shape)  # shape: (num_sample_rays, num_depth_samples + num_importance_samples, 3)

            alpha = self.sigma_to_alpha(sigma=sigma, z_vals=z_vals)  # shape: (num_sample_rays, num_depth_samples + num_importance_samples,)
            weights = self.alpha_to_weights(alpha=alpha)  # shape: (num_sample_rays, num_depth_samples + num_importance_samples,)
            acc_map = self.weights_to_acc_map(weights=weights)  # shape: (num_sample_rays,)
            depth_map = self.weights_to_depth_map(weights=weights, z_vals=z_vals, nears=nears, fars=fars)  # shape: (num_sample_rays,)
            rgb_map = self.weights_to_rgb_map(weights=weights, rgb=rgb)  # shape: (num_sample_rays, 3)
        else:
            acc_map = torch.zeros_like(rays_d[..., 0])  # shape: (num_sample_rays,)
            depth_map = torch.zeros_like(rays_d[..., 0])  # shape: (num_sample_rays,)
            rgb_map = torch.zeros_like(rays_d)  # shape: (num_sample_rays, 3)

        rgb_map = rgb_map + (1 - acc_map).unsqueeze(-1) * self.background_color

        return {
            "rgb_map": rgb_map,
            "depth_map": depth_map,
            "acc_map": acc_map,
        }



# -*- coding: utf-8 -*-
"""
Author: Xiang Yangcheng [https://github.com/Xayah-Hina]
GitHub: https://github.com/IME-lab-Hokudai/EPINF-NeuFlow
Last Modification Date: 2025-07-10
Description: The pytorch implementation of the paper "Efficient Physics Informed Neural Reconstruction of Diverse Fluid Dynamics from Sparse Observations".
License: Mozilla Public License 2.0 (MPL-2.0)
Copyright (c) 2025 Xiang Yangcheng, IME Lab, Hokkaido University, Japan.
"""
from .datasets import PIDatasetAPI, NeRFSyntheticDataset, PINeuFlowDataset
from .epinf import EPINFTrainer
from .models import ModelAPI
from .ngp import NGPTrainer
from .renderers import RendererAPI

__all__ = ['PIDatasetAPI', 'ModelAPI', 'RendererAPI', 'NeRFSyntheticDataset', 'PINeuFlowDataset', 'NGPTrainer', 'EPINFTrainer']



# -*- coding: utf-8 -*-
"""
Author: Xiang Yangcheng [https://github.com/Xayah-Hina]
GitHub: https://github.com/IME-lab-Hokudai/EPINF-NeuFlow
Last Modification Date: 2025-07-10
Description: The pytorch implementation of the paper "Efficient Physics Informed Neural Reconstruction of Diverse Fluid Dynamics from Sparse Observations".
License: Mozilla Public License 2.0 (MPL-2.0)
Copyright (c) 2025 Xiang Yangcheng, IME Lab, Hokkaido University, Japan.
"""
import torch

from .api import Plugin


class TrainSpace(Plugin):
    def __init__(self, resolution: int, name: str = 'tsm'):
        super().__init__(name=name)
        self.register_buffer('trained_grids', torch.zeros((resolution, resolution, resolution), dtype=torch.int32))

        self.res = 256
        coords = torch.linspace(0, 1, steps=self.res + 1)
        centers = (coords[:-1] + coords[1:]) / 2
        x, y, z = torch.meshgrid(centers, centers, centers, indexing='ij')
        self.xyz = torch.stack([x, y, z], dim=-1).view(-1, 3)  # shape: (res^3, 1)

    def to(self, device):
        super().to(device)
        self.trained_grids = self.trained_grids.to(device=device)
        self.xyz = self.xyz.to(device=device)

    def update(self, xyz, index):
        assert xyz.max() <= 1.0 and xyz.min() >= 0.0, "Normalized xyz coordinates should be in [0, 1] range."
        assert index < 32, "Index must be less than 32."
        res = self.trained_grids.shape[0]  # Assuming trained_grids is a 4D tensor with shape (1, res, res, res)
        idx = (xyz * res).long().clamp(0, res - 1)
        x, y, z = idx[:, 0], idx[:, 1], idx[:, 2]

        bitmask = 1 << index
        before = self.trained_grids[x, y, z]
        changed = (before & bitmask) == 0
        self.trained_grids[x, y, z] = before | bitmask
        num_changed = changed.sum().item()
        return num_changed

    def exclusive_mask(self, xyz: torch.Tensor):
        pass

    def exclusive_grid(self):
        nonzero_mask = self.trained_grids != 0  # shape: (N, N, N)，bool
        # 初始化一个全 False 的 mask
        mask = torch.zeros_like(self.trained_grids, dtype=torch.bool)

        # 获取所有非零元素的索引
        indices = nonzero_mask.nonzero(as_tuple=False)  # shape: (M, 3)
        values = self.trained_grids[nonzero_mask]  # shape: (M,)

        # 判断是否是2的幂
        power_of_two_mask = (values & (values - 1)) == 0  # shape: (M,)

        # 只保留满足条件的位置
        selected_indices = indices[power_of_two_mask]  # shape: (K, 3)

        # 将这些位置设为 True
        mask[selected_indices[:, 0], selected_indices[:, 1], selected_indices[:, 2]] = True

        return mask  # shape: (N, N, N)，bool

    def exclusive_n_mask(self, xyz: torch.Tensor, n: int):
        res = self.trained_grids.shape[0]  # Assuming trained_grids is a 4D tensor with shape (1, res, res, res)
        idx = (xyz * res).long().clamp(0, res - 1)
        x, y, z = idx[:, 0], idx[:, 1], idx[:, 2]

        values = self.trained_grids[x, y, z]
        bit_counts = self.popcount32(values)
        mask = (bit_counts == n)
        return mask

    def exclusive_n_grid(self, n: int):
        assert 1 <= n <= 32
        grid = self.trained_grids  # shape: (res, res, res), dtype=torch.int32
        bit_counts = self.popcount32(grid)  # shape: (res, res, res), dtype=torch.int32
        mask = (bit_counts == n)  # shape: (res, res, res), dtype=torch.bool
        return mask

    def overlapped_grid(self):
        grid = self.trained_grids  # shape: (res, res, res), dtype=torch.int32
        bit_counts = self.popcount32(grid)  # 每个 voxel 有多少个 bit 被设置
        max_bits = bit_counts.max()  # 当前最大 bit 数
        mask = (bit_counts == max_bits)  # 哪些 voxel 的 bit 数是最大值
        return mask

    def empty_grid(self):
        return self.trained_grids == 0

    def occupied_grid(self):
        return self.trained_grids != 0

    @staticmethod
    def popcount32(x: torch.Tensor) -> torch.Tensor:
        """
        Bit-parallel population count for int32 tensor.
        Input: x (torch.int32 tensor)
        Output: int32 tensor of same shape, each value is the number of 1s in x[i].
        """
        x = x - ((x >> 1) & 0x55555555)
        x = (x & 0x33333333) + ((x >> 2) & 0x33333333)
        x = (x + (x >> 4)) & 0x0F0F0F0F
        x = (x * 0x01010101) >> 24
        return x



# -*- coding: utf-8 -*-
"""
Author: Xiang Yangcheng [https://github.com/Xayah-Hina]
GitHub: https://github.com/IME-lab-Hokudai/EPINF-NeuFlow
Last Modification Date: 2025-07-10
Description: The pytorch implementation of the paper "Efficient Physics Informed Neural Reconstruction of Diverse Fluid Dynamics from Sparse Observations".
License: Mozilla Public License 2.0 (MPL-2.0)
Copyright (c) 2025 Xiang Yangcheng, IME Lab, Hokkaido University, Japan.
"""
import dataclasses
import math

import tinycudann
import torch

from .api import ModelAPI


class EPINFStaticModel(ModelAPI):
    @dataclasses.dataclass
    class Config:
        num_layers_sigma: int = dataclasses.field(default=3, metadata={'help': 'number of layers for the sigma network'})
        hidden_dim_sigma: int = dataclasses.field(default=64, metadata={'help': 'hidden dimension for the sigma network'})
        num_layers_rgb: int = dataclasses.field(default=3, metadata={'help': 'number of layers for the RGB network'})
        hidden_dim_rgb: int = dataclasses.field(default=64, metadata={'help': 'hidden dimension for the RGB network'})
        geo_feat_dim: int = dataclasses.field(default=32, metadata={'help': 'geometric feature dimension'})
        num_layers_bg: int = dataclasses.field(default=3, metadata={'help': 'number of layers for the background network'})
        hidden_dim_bg: int = dataclasses.field(default=64, metadata={'help': 'hidden dimension for the background network'})
        use_background: bool = dataclasses.field(default=False, metadata={'help': 'whether to use background in the dataset'})

    def __init__(self, cfg: Config):
        super().__init__(cfg=cfg)

        bound = 1
        self.encoder_sigma = tinycudann.Encoding(
            n_input_dims=3,
            encoding_config={
                "otype": "HashGrid",
                "n_levels": 16,
                "n_features_per_level": 2,
                "log2_hashmap_size": 19,
                "base_resolution": 16,
                "per_level_scale": 2 ** (math.log2(2048 * bound / 16) / (16 - 1)),
            },
        )
        self.sigma_net = tinycudann.Network(
            n_input_dims=int(self.encoder_sigma.n_output_dims),
            n_output_dims=1 + cfg.geo_feat_dim,
            network_config={
                "otype": "FullyFusedMLP",
                "activation": "ReLU",
                "output_activation": "None",
                "n_neurons": cfg.hidden_dim_sigma,
                "n_hidden_layers": cfg.num_layers_sigma - 1,
            },
        )

        self.encoder_dir = tinycudann.Encoding(
            n_input_dims=3,
            encoding_config={
                "otype": "SphericalHarmonics",
                "degree": 4,
            },
        )
        self.color_net = tinycudann.Network(
            n_input_dims=int(self.encoder_dir.n_output_dims) + cfg.geo_feat_dim,
            n_output_dims=3,
            network_config={
                "otype": "FullyFusedMLP",
                "activation": "ReLU",
                "output_activation": "None",
                "n_neurons": cfg.hidden_dim_rgb,
                "n_hidden_layers": cfg.num_layers_rgb - 1,
            },
        )

        if cfg.use_background:
            self.encoder_bg = tinycudann.Encoding(
                n_input_dims=2,
                encoding_config={
                    "otype": "HashGrid",
                    "n_levels": 4,
                    "n_features_per_level": 2,
                    "log2_hashmap_size": 19,
                    "base_resolution": 16,
                    "per_level_scale": 2 ** (math.log2(2048 * bound / 4) / (4 - 1)),
                },
            )
            self.bg_net = tinycudann.Network(
                n_input_dims=int(self.encoder_dir.n_output_dims) + int(self.encoder_bg.n_output_dims),
                n_output_dims=3,
                network_config={
                    "otype": "FullyFusedMLP",
                    "activation": "ReLU",
                    "output_activation": "None",
                    "n_neurons": cfg.hidden_dim_bg,
                    "n_hidden_layers": cfg.num_layers_bg - 1,
                },
            )

    def sigma(self, xyz):
        xyz_encoded = self.encoder_sigma(xyz)
        h = self.sigma_net(xyz_encoded)
        sigma = torch.nn.functional.relu(h[..., :1])
        geo_feat = h[..., 1:]
        return sigma, geo_feat

    def rgb(self, xyz, view_dirs, geo_feat, cond=None):
        view_dirs_encoded = self.encoder_dir(view_dirs)
        h = self.color_net(torch.cat([view_dirs_encoded, geo_feat], dim=-1))
        rgb = torch.sigmoid(h)
        return rgb

    def background(self, sph, view_dirs):
        sph_encoded = self.encoder_bg(sph)
        view_dirs_encoded = self.encoder_dir(view_dirs)
        h = self.bg_net(torch.cat([view_dirs_encoded, sph_encoded], dim=-1))
        rgb_bg = torch.sigmoid(h)
        return rgb_bg


class EPINFDynamicModel(ModelAPI):
    @dataclasses.dataclass
    class Config:
        num_layers_sigma: int = dataclasses.field(default=3, metadata={'help': 'number of layers for the sigma network'})
        hidden_dim_sigma: int = dataclasses.field(default=64, metadata={'help': 'hidden dimension for the sigma network'})
        num_layers_rgb: int = dataclasses.field(default=3, metadata={'help': 'number of layers for the RGB network'})
        hidden_dim_rgb: int = dataclasses.field(default=64, metadata={'help': 'hidden dimension for the RGB network'})
        geo_feat_dim: int = dataclasses.field(default=32, metadata={'help': 'geometric feature dimension'})

    def __init__(self, cfg: Config):
        super().__init__(cfg=cfg)

        bound = 1
        self.encoder = tinycudann.Encoding(
            n_input_dims=4,
            encoding_config={
                "otype": "HashGrid",
                "n_levels": 16,
                "n_features_per_level": 2,
                "log2_hashmap_size": 19,
                "base_resolution": 16,
                "per_level_scale": 2 ** (math.log2(2048 * bound / 16) / (16 - 1)),
            },
        )

        self.sigma_net = tinycudann.Network(
            n_input_dims=int(self.encoder.n_output_dims),
            n_output_dims=1 + cfg.geo_feat_dim,
            network_config={
                "otype": "FullyFusedMLP",
                "activation": "ReLU",
                "output_activation": "None",
                "n_neurons": cfg.hidden_dim_sigma,
                "n_hidden_layers": cfg.num_layers_sigma - 1,
            },
        )

        self.encoder_dir = tinycudann.Encoding(
            n_input_dims=3,
            encoding_config={
                "otype": "SphericalHarmonics",
                "degree": 4,
            },
        )

        self.color_net = tinycudann.Network(
            n_input_dims=int(self.encoder_dir.n_output_dims) + cfg.geo_feat_dim,
            n_output_dims=3,
            network_config={
                "otype": "FullyFusedMLP",
                "activation": "ReLU",
                "output_activation": "None",
                "n_neurons": cfg.hidden_dim_rgb,
                "n_hidden_layers": cfg.num_layers_rgb - 1,
            },
        )

        self._time = None

    @property
    def time(self):
        assert self._time is not None
        assert isinstance(self._time, torch.Tensor)
        assert self._time.shape[0] == 1
        return self._time

    def set_time(self, time):
        self._time = time

    def sigma(self, xyz):
        xyzt = torch.cat([xyz, self.time.unsqueeze(0).expand_as(xyz[..., :1])], dim=-1)
        xyzt_encoded = self.encoder(xyzt)
        h = self.sigma_net(xyzt_encoded)
        sigma = torch.nn.functional.relu(h[..., :1])
        geo_feat = h[..., 1:]
        return sigma, geo_feat

    def rgb(self, xyz, view_dirs, geo_feat, cond=None):
        view_dirs_encoded = self.encoder_dir(view_dirs)
        h = self.color_net(torch.cat([view_dirs_encoded, geo_feat], dim=-1))
        rgb = torch.sigmoid(h)
        return rgb

    def background(self, sph, view_dirs):
        raise NotImplementedError("Background rendering is not implemented for dynamic models.")


class EPINFDynamicModelDoubleGradable(EPINFDynamicModel):
    def __init__(self, cfg: EPINFDynamicModel.Config):
        super().__init__(cfg)

        # Override the sigma network to allow double gradient computation
        sigma_net = []
        for l in range(cfg.num_layers_sigma):
            if l == 0:
                in_dim = self.encoder.n_output_dims
            else:
                in_dim = cfg.hidden_dim_sigma
            if l == cfg.num_layers_sigma - 1:
                out_dim = 1 + cfg.geo_feat_dim
            else:
                out_dim = cfg.hidden_dim_sigma
            sigma_net.append(torch.nn.Linear(in_dim, out_dim, bias=False))
        self.sigma_net = torch.nn.ModuleList(sigma_net)

    def sigma(self, xyz):
        xyzt = torch.cat([xyz, self.time.unsqueeze(0).expand_as(xyz[..., :1])], dim=-1)
        if self.training:
            xyzt.requires_grad = True
        xyzt_encoded = self.encoder(xyzt)

        h = xyzt_encoded
        for l in range(self.cfg.num_layers_sigma):
            h = self.sigma_net[l](h)
            if l != self.cfg.num_layers_sigma - 1:
                h = torch.nn.functional.relu(h, inplace=True)

        sigma = torch.nn.functional.relu(h[..., :1])
        geo_feat = h[..., 1:]
        return sigma, geo_feat, xyzt_encoded, xyzt

    def fn(self, hidden):
        h = hidden
        for l in range(self.cfg.num_layers_sigma):
            h = self.sigma_net[l](h)
            if l != self.cfg.num_layers_sigma - 1:
                h = torch.nn.functional.relu(h, inplace=True)
        return torch.nn.functional.relu(h[..., :1])


class EPINFVelocityModel(ModelAPI):
    @dataclasses.dataclass
    class Config:
        num_layers_vel_sigma: int = dataclasses.field(default=3, metadata={'help': 'number of layers for the sigma network'})
        hidden_dim_vel_sigma: int = dataclasses.field(default=64, metadata={'help': 'hidden dimension for the sigma network'})

    def __init__(self, cfg: Config):
        super().__init__(cfg=cfg)

        bound = 1
        self.encoder_vel = tinycudann.Encoding(
            n_input_dims=4,
            encoding_config={
                "otype": "HashGrid",
                "n_levels": 16,
                "n_features_per_level": 2,
                "log2_hashmap_size": 19,
                "base_resolution": 16,
                "per_level_scale": 2 ** (math.log2(2048 * bound / 16) / (16 - 1)),
            },
        )

        self.sigma_vel_net = tinycudann.Network(
            n_input_dims=int(self.encoder_vel.n_output_dims),
            n_output_dims=3,
            network_config={
                "otype": "FullyFusedMLP",
                "activation": "ReLU",
                "output_activation": "None",
                "n_neurons": cfg.hidden_dim_vel_sigma,
                "n_hidden_layers": cfg.num_layers_vel_sigma - 1,
            },
        )

        self._time = None

    @property
    def time(self):
        assert self._time is not None
        assert isinstance(self._time, torch.Tensor)
        assert self._time.shape[0] == 1
        return self._time

    def set_time(self, time):
        self._time = time

    def vel(self, xyz):
        return self.sigma(xyz)

    def sigma_grad(self, xyzt_grad):
        xyzt_encoded = self.encoder_vel(xyzt_grad)
        velocity = self.sigma_vel_net(xyzt_encoded)
        return velocity

    def sigma(self, xyz):
        xyzt = torch.cat([xyz, self.time.unsqueeze(0).expand_as(xyz[..., :1])], dim=-1)
        xyzt_encoded = self.encoder_vel(xyzt)
        velocity = self.sigma_vel_net(xyzt_encoded)
        return velocity


class EPINFHybridModel(ModelAPI):
    @dataclasses.dataclass
    class Config:
        static: EPINFStaticModel.Config = dataclasses.field(default_factory=EPINFStaticModel.Config, metadata={'help': 'Configuration for the static model'})
        dynamic: EPINFDynamicModel.Config = dataclasses.field(default_factory=EPINFDynamicModel.Config, metadata={'help': 'Configuration for the dynamic model'})
        vel: EPINFVelocityModel.Config = dataclasses.field(default_factory=EPINFVelocityModel.Config, metadata={'help': 'Configuration for the velocity model'})

        legacy: bool = dataclasses.field(default=False, metadata={'help': 'Whether to use legacy mode for the hybrid model'})

    def __init__(self, cfg: Config):
        super().__init__(cfg=cfg)
        self.static_renderer: EPINFStaticModel = EPINFStaticModel(cfg=cfg.static)
        self.dynamic_renderer: EPINFDynamicModel | EPINFDynamicModelDoubleGradable = EPINFDynamicModelDoubleGradable(cfg=cfg.dynamic) if cfg.legacy else EPINFDynamicModel(cfg=cfg.dynamic)
        self.vel_renderer: EPINFVelocityModel = EPINFVelocityModel(cfg=cfg.vel)

    @property
    def time(self):
        return self.dynamic_renderer.time

    def set_time(self, time):
        self.dynamic_renderer.set_time(time)
        self.vel_renderer.set_time(time)

    def sigma(self, xyz):
        ret_sta = self.static_renderer.sigma(xyz)
        ret_dyn = self.dynamic_renderer.sigma(xyz)
        sigma_static, geo_feat_static = ret_sta[0], ret_sta[1]
        sigma_dynamic, geo_feat_dynamic = ret_dyn[0], ret_dyn[1]
        return sigma_static + sigma_dynamic, geo_feat_static + geo_feat_dynamic

    def rgb(self, xyz, view_dirs, geo_feat, cond=None):
        raise NotImplementedError("Hybrid model does not support direct RGB rendering. Use static or dynamic renderer instead.")

    def vel(self, xyz):
        return self.vel_renderer.vel(xyz)

    def background(self, sph, view_dirs):
        raise NotImplementedError("Background rendering is not implemented for hybrid models. Use static or dynamic renderer instead.")



# -*- coding: utf-8 -*-
"""
Author: Xiang Yangcheng [https://github.com/Xayah-Hina]
GitHub: https://github.com/IME-lab-Hokudai/EPINF-NeuFlow
Last Modification Date: 2025-07-10
Description: The pytorch implementation of the paper "Efficient Physics Informed Neural Reconstruction of Diverse Fluid Dynamics from Sparse Observations".
License: Mozilla Public License 2.0 (MPL-2.0)
Copyright (c) 2025 Xiang Yangcheng, IME Lab, Hokkaido University, Japan.
"""
import dataclasses
import glob
import os
import subprocess
import tempfile
import textwrap

import lightning
import numpy as np
import rich
import rich.progress
import torch
import torchmetrics
import torchvision.utils

from .datasets import PIDatasetAPI, NeRFSyntheticDataset
from .models import NGPModel
from .plugins import PluginLoss
from .renderers import NGPRenderer


class NGPTrainer(lightning.LightningModule):
    @dataclasses.dataclass
    class NGP:
        dataset: NeRFSyntheticDataset.Config = dataclasses.field(default_factory=NeRFSyntheticDataset.Config, metadata={'help': 'NeRF synthetic dataset configuration'})
        model: NGPModel.Config = dataclasses.field(default_factory=NGPModel.Config, metadata={'help': 'model configuration'})
        renderer: NGPRenderer.Config = dataclasses.field(default_factory=NGPRenderer.Config, metadata={'help': 'renderer configuration'})

        # general settings
        gpu_id: int = dataclasses.field(default=0, metadata={'help': 'GPU ID to use for training'})
        gui: bool = dataclasses.field(default=False, metadata={'help': 'whether to use GUI for rendering'})
        ckpt: bool = dataclasses.field(default=False, metadata={'help': 'checkpoint path to load for testing'})

        # training settings
        epochs: int = dataclasses.field(default=100, metadata={'help': 'number of epochs to train the model'})
        lrate: float = dataclasses.field(default=1e-3, metadata={'help': 'learning rate for the optimizer'})
        val_skip: int = dataclasses.field(default=10, metadata={'help': 'number of epochs to skip for validation'})
        save_val_map: bool = dataclasses.field(default=True, metadata={'help': 'whether to output validation images'})
        export_model_only: bool = dataclasses.field(default=False, metadata={'help': 'whether to export model only'})

        # test settings
        test: bool = dataclasses.field(default=False, metadata={'help': 'test mode, load a checkpoint and test the model'})
        export: bool = dataclasses.field(default=False, metadata={'help': 'whether to export fields'})
        export_houdini: bool = dataclasses.field(default=False, metadata={'help': 'whether to export fields'})
        fields: bool = dataclasses.field(default=True, metadata={'help': 'whether to export fields, defaults to True when export or export_houdini is True'})

    def __init__(self, cfg: NGP):
        super().__init__()
        self.cfg = cfg
        self.save_hyperparameters()
        self.model = NGPModel(cfg=cfg.model)
        self.renderer = NGPRenderer(cfg=cfg.renderer)
        self.psnr = torchmetrics.image.PeakSignalNoiseRatio(data_range=1.0)
        self.ssim = torchmetrics.image.StructuralSimilarityIndexMeasure(data_range=1.0)
        self.lpips = torchmetrics.image.lpip.LearnedPerceptualImagePatchSimilarity(net_type='vgg')

    def on_fit_start(self):
        dataset: PIDatasetAPI = self.trainer.datamodule
        self.renderer.set_render_resolution(width=dataset.width, height=dataset.height)
        rich.print(f"[bold green]>>> [Custom]: Render resolution set to {dataset.width}x{dataset.height} <<<[/bold green]")
        self.renderer.set_min_near_far(min_near=dataset.near_std, max_far=dataset.far_std)
        rich.print(f"[bold green]>>> [Custom]: Near and far std set to {dataset.near_std} and {dataset.far_std} <<<[/bold green]")
        self.renderer.set_aabb_std(dataset.aabb_std)
        rich.print(f"[bold green]>>> [Custom]: AABB std set to {dataset.aabb_std} <<<[/bold green]")
        self.renderer.set_bound_std(dataset.bound_std)
        rich.print(f"[bold green]>>> [Custom]: Bound std set to {dataset.bound_std} <<<[/bold green]")
        self.renderer.set_background_color(dataset.background_color)
        rich.print(f"[bold green]>>> [Custom]: Background color set to {dataset.background_color} <<<[/bold green]")
        self.renderer.to(device=self.device)
        rich.print(f"[bold green]>>> [Custom]: Renderer moved to device {self.device} <<<[/bold green]")

    def on_train_batch_start(self, batch, batch_idx):
        for plugin in self.renderer.plugins:
            plugin.on_iter_start()

    def training_step(self, batch, batch_idx):
        result_maps, pixels, pixels_mask = self.renderer.forward(
            model=self.model,
            poses=batch['poses'],
            focals=batch['focals'],
            images=batch['images'],
            pose_indices=batch.get('pose_indices', None),
            images_masks=batch.get('images_masks', None),
        )

        rgb_map = result_maps['rgb_map']
        img_loss = 10000 * torch.nn.functional.mse_loss(rgb_map, pixels[..., :3])
        self.log("img_loss", img_loss, on_step=True, on_epoch=False, prog_bar=True)
        loss = img_loss

        for plugin_loss in filter(lambda v: isinstance(v, PluginLoss), self.renderer.plugins):
            loss += plugin_loss.weight * plugin_loss.loss
            self.log(f"{plugin_loss.name}", plugin_loss.loss, on_step=True, on_epoch=False, prog_bar=True)

        self.log("loss", loss, on_step=True, on_epoch=False, prog_bar=True)
        return loss

    def on_validation_start(self):
        if self.cfg.save_val_map:
            os.makedirs(os.path.join(self.logger.experiment.dir, 'validation_images'), exist_ok=True)

    def validation_step(self, batch, batch_idx):
        if batch_idx % self.cfg.val_skip != 0:
            return None

        result_maps, pixels, pixels_mask = self.renderer.forward(
            model=self.model,
            poses=batch['poses'],
            focals=batch['focals'],
            images=batch['images'],
            pose_indices=batch.get('pose_indices', None),
            images_masks=batch.get('images_masks', None),
        )

        val_loss = torch.nn.functional.mse_loss(result_maps['rgb_map'], pixels[..., :3])
        self.log("val_loss", val_loss, on_step=False, on_epoch=True, prog_bar=True)
        psnr = self.psnr(result_maps['rgb_map'], pixels[..., :3])
        self.log("val_psnr", psnr, on_step=False, on_epoch=True, prog_bar=True)
        ssim = self.ssim(result_maps['rgb_map'].permute(0, 3, 1, 2), pixels[..., :3].permute(0, 3, 1, 2))  # (N, C, H, W)
        self.log("val_ssim", ssim, on_step=False, on_epoch=True, prog_bar=True)
        lpips = self.lpips(result_maps['rgb_map'].permute(0, 3, 1, 2), pixels[..., :3].permute(0, 3, 1, 2))
        self.log("val_lpips", lpips, on_step=False, on_epoch=True, prog_bar=True)

        if self.cfg.save_val_map:
            final_map = self.generate_map(result_maps, pixels, batch.get('images_highlighted', None), pixels_mask)
            torchvision.utils.save_image(final_map.permute(2, 0, 1), os.path.join(self.logger.experiment.dir, 'validation_images', f'{self.current_epoch}_{batch_idx}.png'))

        return val_loss

    def on_test_start(self):
        os.makedirs(os.path.join(self.logger.experiment.dir, 'test_images'), exist_ok=True)

    def test_step(self, batch, batch_idx):
        result_maps, pixels, pixels_mask = self.renderer.forward(
            model=self.model,
            poses=batch['poses'],
            focals=batch['focals'],
            images=batch.get('images', None),
            pose_indices=batch.get('pose_indices', None),
            images_masks=batch.get('images_masks', None),
        )

        final_map = self.generate_map(result_maps, pixels, batch.get('images_highlighted', None), pixels_mask)

        psnr = self.psnr(result_maps['rgb_map'], pixels[..., :3])
        self.log("test_psnr", psnr, on_step=False, on_epoch=True, prog_bar=True)
        ssim = self.ssim(result_maps['rgb_map'].permute(0, 3, 1, 2), pixels[..., :3].permute(0, 3, 1, 2))  # (N, C, H, W)
        self.log("test_ssim", ssim, on_step=False, on_epoch=True, prog_bar=True)
        lpips = self.lpips(result_maps['rgb_map'].permute(0, 3, 1, 2), pixels[..., :3].permute(0, 3, 1, 2))
        self.log("test_lpips", lpips, on_step=False, on_epoch=True, prog_bar=True)
        torchvision.utils.save_image(final_map.permute(2, 0, 1), os.path.join(self.logger.experiment.dir, 'test_images', f'{batch_idx}.png'))

    def on_test_end(self):
        image_dir = os.path.join(self.logger.experiment.dir, 'test_images')
        video_path = os.path.join(self.logger.experiment.dir, 'test_images', 'test_video.mp4')

        image_paths = sorted(
            glob.glob(os.path.join(image_dir, '*.png')),
            key=lambda x: int(os.path.basename(x).split('_')[-1].split('.')[0])
        )

        frames = []
        for path in image_paths:
            img = torchvision.io.read_image(path)  # (C, H, W), dtype=torch.uint8, range [0,255]
            if img.shape[0] == 1:
                img = img.expand(3, -1, -1)
            frames.append(img.permute(1, 2, 0))  # (H, W, C)

        video_tensor = torch.stack(frames, dim=0)  # (T, H, W, 3)
        torchvision.io.write_video(video_path, video_tensor, fps=24)
        rich.print(f"[bold green]✅ Video saved at[/bold green] [cyan]{video_path}[/cyan] with {len(frames)} frames")

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(params=self.parameters(), lr=self.cfg.lrate, eps=1e-15)
        scheduler = {
            "scheduler": torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, mode='min', factor=0.9, patience=10, min_lr=1e-5),
            "monitor": "loss",  # what to monitor for the scheduler
            "interval": "epoch",  # when to step the scheduler
            "frequency": 1,
        }
        return {"optimizer": optimizer, "lr_scheduler": scheduler}

    @staticmethod
    def generate_map(result_maps, pixels, pixels_highlighted, pixels_mask):
        rgb_map_column = torch.cat([rgb for rgb in result_maps['rgb_map']], dim=1)
        acc_map_column = torch.cat([rgb for rgb in result_maps['acc_map']], dim=1) if 'acc_map' in result_maps else None
        depth_map_column = torch.cat([rgb for rgb in result_maps['depth_map']], dim=1) if 'depth_map' in result_maps else None

        pixels_column = torch.cat([pixel for pixel in pixels[..., :3]], dim=1) if pixels is not None else None
        pixels_highlighted_column = torch.cat([pixel for pixel in pixels_highlighted[..., :3]], dim=1) if pixels_highlighted is not None else None
        pixels_mask_column = torch.cat([mask for mask in pixels_mask.unsqueeze(-1)], dim=1) if pixels_mask is not None else None

        final_map = rgb_map_column
        if acc_map_column is not None:
            final_map = torch.cat([final_map, acc_map_column.expand_as(rgb_map_column)], dim=1)
        if depth_map_column is not None:
            final_map = torch.cat([final_map, depth_map_column.expand_as(rgb_map_column)], dim=1)
        if pixels_mask_column is not None:
            final_map = torch.cat([pixels_mask_column.expand_as(rgb_map_column), final_map], dim=1)
        if pixels_highlighted_column is not None:
            final_map = torch.cat([pixels_highlighted_column, final_map], dim=1)
        if pixels_column is not None:
            final_map = torch.cat([pixels_column, final_map], dim=1)

        return final_map

    def export_explicit_fields(self, output_dir, hython_path):
        aabb_std = self.renderer.aabb_std.cpu()  # [low_x, low_y, low_z, high_x, high_y, high_z]
        bound_std = self.renderer.bound_std.cpu()  # [low_x, low_y, low_z, high_x, high_y, high_z]

        weightx = (aabb_std[3] - aabb_std[0]).item() / 2
        weighty = (aabb_std[4] - aabb_std[1]).item() / 2
        weightz = (aabb_std[5] - aabb_std[2]).item() / 2

        min_bound = bound_std[:3]  # [low_x, low_y, low_z]
        max_bound = bound_std[3:]  # [high_x, high_y, high_z]

        resx, resy, resz = int(256 * weightx), int(256 * weighty), int(256 * weightz)
        coords_x = torch.linspace(0, 1, steps=resx + 1)
        centers_x = (coords_x[:-1] + coords_x[1:]) / 2
        coords_y = torch.linspace(0, 1, steps=resy + 1)
        centers_y = (coords_y[:-1] + coords_y[1:]) / 2
        coords_z = torch.linspace(0, 1, steps=resz + 1)
        centers_z = (coords_z[:-1] + coords_z[1:]) / 2
        x, y, z = torch.meshgrid(centers_x, centers_y, centers_z, indexing='ij')
        xyz = torch.stack([x, y, z], dim=-1).view(-1, 3)
        mask_inside = ((xyz >= min_bound) & (xyz <= max_bound)).all(dim=-1)
        xyz = xyz.to(torch.float16 if self.cfg.dataset.fp16 else torch.float32).to(self.device)

        with torch.no_grad():
            with torch.amp.autocast('cuda', enabled=True):
                self.eval()
                with rich.progress.Progress() as progress:
                    task = progress.add_task("[green]Exporting density fields...", total=xyz.shape[0])
                    chunk = 1024 * 8
                    sigma = []
                    for start_idx in range(0, xyz.shape[0], chunk):
                        ret = self.model.sigma(xyz[start_idx:start_idx + chunk])
                        _sigma = ret[0]
                        sigma.append(_sigma.detach().cpu())
                        progress.update(task, advance=(min(start_idx + chunk, xyz.shape[0]) - start_idx))
                    sigma = torch.cat(sigma, dim=0)
                    sigma[~mask_inside] = 0.0
                    sigma = sigma.reshape(resx, resy, resz, 1)

                if hython_path is not None:
                    def execute_template_script(script_content):
                        with tempfile.NamedTemporaryFile(suffix=".py", delete=False, mode="w", encoding="utf-8") as f:
                            temp_script_path = f.name
                            f.write(textwrap.dedent(script_content))
                        try:
                            subprocess.run([hython_path, temp_script_path], check=True)
                        except subprocess.CalledProcessError as e:
                            raise RuntimeError(f"Failed to execute Houdini script: {e}") from e
                        finally:
                            os.remove(temp_script_path)

                    with tempfile.NamedTemporaryFile(suffix=".npy", delete=False) as f:
                        temp_file = f.name
                    np.save(temp_file, sigma.numpy())
                    houdini_code = f'''
                        import hou
                        import numpy as np
                        import os
                        nparray = np.load(r"{temp_file}")
                        resx, resy, resz = nparray.shape[0:3]
                        geo = hou.Geometry()
                        name_attrib = geo.addAttrib(hou.attribType.Prim, "name", "default")
                        vol = geo.createVolume(resx, resy, resz, hou.BoundingBox(0.0, 0.0, 0.0, 1.0, 1.0, 1.0))
                        vol.setAttribValue(name_attrib, "density")
                        vol.setAllVoxels(nparray.flatten().tolist())
                        os.makedirs(r"{output_dir}", exist_ok=True)
                        out_file = os.path.join(r"{output_dir}", f"sigma.bgeo.sc")
                        geo.saveToFile(out_file)
                        '''
                    try:
                        execute_template_script(script_content=houdini_code)
                        gplay_exe = os.path.join(os.path.dirname(hython_path), 'gplay.exe')
                        sigma_file = os.path.join(output_dir, f"sigma.bgeo.sc")
                        rich.print(f"[bold green]Executing[/bold green] [cyan]{gplay_exe} {sigma_file}[/cyan]")
                        subprocess.run([gplay_exe, sigma_file], check=True)
                    except subprocess.CalledProcessError as e:
                        raise RuntimeError(f"Failed to execute Houdini script: {e}") from e
                    finally:
                        os.remove(temp_file)
                else:
                    torch.save(sigma, os.path.join(output_dir, f'sigma.pt'))
                    rich.print(f"[bold green]✅ Density field saved at[/bold green] [cyan]{os.path.join(output_dir, f'sigma.pt')}[/cyan] with shape [magenta]{sigma.shape}[/magenta]")



# -*- coding: utf-8 -*-
"""
Author: Xiang Yangcheng [https://github.com/Xayah-Hina]
GitHub: https://github.com/IME-lab-Hokudai/EPINF-NeuFlow
Last Modification Date: 2025-07-10
Description: The pytorch implementation of the paper "Efficient Physics Informed Neural Reconstruction of Diverse Fluid Dynamics from Sparse Observations".
License: Mozilla Public License 2.0 (MPL-2.0)
Copyright (c) 2025 Xiang Yangcheng, IME Lab, Hokkaido University, Japan.
"""
import os

import cv2
import numpy as np
import rich


def extract_full_foreground_mask(frame, bg_color, color_thresh, dilation_pixels):
    """
    从指定颜色背景图中提取最大前景区域（物体+流体），返回红色叠加图和二值mask（只保留最大连通域）

    参数：
    - frame: 输入图像，np.uint8, shape=(H, W, 3)
    - bg_color: 背景颜色 (R, G, B)
    - color_thresh: 与背景色的欧氏距离阈值
    - dilation_pixels: mask膨胀范围（像素）
    """
    # 1. 计算与背景颜色的距离
    diff = frame.astype(np.int16) - np.array(bg_color, dtype=np.int16)[None, None, :]
    dist = np.linalg.norm(diff, axis=2).astype(np.uint8)

    # 2. 阈值处理得到初始前景
    _, binary = cv2.threshold(dist, color_thresh, 255, cv2.THRESH_BINARY)

    # 3. 膨胀操作
    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (2 * dilation_pixels + 1, 2 * dilation_pixels + 1))
    dilated_mask = cv2.dilate(binary, kernel)

    # 4. 连通域分析：只保留最大区域
    num_labels, labels, stats, _ = cv2.connectedComponentsWithStats(dilated_mask, connectivity=8)

    # 找到最大区域（跳过背景 label=0）
    max_label = 1 + np.argmax(stats[1:, cv2.CC_STAT_AREA])

    # 构造 mask，仅保留最大连通域
    largest_mask = np.zeros_like(dilated_mask)
    largest_mask[labels == max_label] = 255

    # 5. 红色高亮叠加
    alpha_mask = (largest_mask / 255.0)[..., None]
    red = np.full_like(frame, (0, 0, 255), dtype=np.uint8)
    blended = frame.astype(np.float32) * (1 - 0.4 * alpha_mask) + red.astype(np.float32) * (0.4 * alpha_mask)
    blended = blended.astype(np.uint8)

    return blended, largest_mask


def process_video_full_mask(input_path, background_color, color_thresh=20, dilation_pixels=8):
    cap = cv2.VideoCapture(input_path)
    fourcc = cv2.VideoWriter_fourcc(*'mp4v')
    fps = cap.get(cv2.CAP_PROP_FPS)
    width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))

    # 输出路径
    dirname, basename = os.path.split(input_path)
    name, ext = os.path.splitext(basename)
    out_color_path = os.path.join(dirname, f"{name}_highlighted{ext}")
    out_mask_path = os.path.join(dirname, f"{name}_mask{ext}")

    out_color = cv2.VideoWriter(out_color_path, fourcc, fps, (width, height))
    out_mask = cv2.VideoWriter(out_mask_path, fourcc, fps, (width, height), isColor=False)

    while True:
        ret, frame = cap.read()
        if not ret:
            break

        highlighted, binary_mask = extract_full_foreground_mask(frame=frame, bg_color=background_color, color_thresh=color_thresh, dilation_pixels=dilation_pixels)

        out_color.write(highlighted)
        out_mask.write(binary_mask)

    cap.release()
    out_color.release()
    out_mask.release()
    rich.print(f"[bold green]Processed video:[/bold green] {input_path}")


if __name__ == "__main__":
    for video in ['train00.mp4', 'train01.mp4', 'train02.mp4', 'train03.mp4', 'train04.mp4']:
        process_video_full_mask(
            input_path="../../data/pi_neuflow/scalar/" + video,
            background_color=(0, 0, 0),  # 黑色背景
            color_thresh=50,  # 与黑色背景的距离阈值
            dilation_pixels=30,  # 膨胀范围
        )
    for video in ['train00.mp4', 'train01.mp4', 'train02.mp4', 'train03.mp4', 'train04.mp4']:
        process_video_full_mask(
            input_path="../../data/pi_neuflow/sphere/" + video,
            background_color=(255, 255, 255),  # 白色背景
            color_thresh=20,  # 与白色背景的距离阈值
            dilation_pixels=8,  # 膨胀范围
        )
    for video in ['train00.mp4', 'train01.mp4', 'train02.mp4', 'train03.mp4', 'train04.mp4', 'train05.mp4', 'train06.mp4']:
        process_video_full_mask(
            input_path="../../data/pi_neuflow/game/" + video,
            background_color=(0, 0, 0),  # 黑色背景
            color_thresh=20,  # 与黑色背景的距离阈值
            dilation_pixels=20,  # 膨胀范围
        )
    for video in ['Torch Simple cam1.mp4', 'Torch Simple cam2.mp4', 'Torch Simple cam3.mp4', 'Torch Simple cam4.mp4', 'Torch Simple cam5.mp4']:
        process_video_full_mask(
            input_path="../../data/pi_neuflow/torch/" + video,
            background_color=(0, 0, 0),  # 黑色背景
            color_thresh=20,  # 与黑色背景的距离阈值
            dilation_pixels=20,  # 膨胀范围
        )



# -*- coding: utf-8 -*-
"""
Author: Xiang Yangcheng [https://github.com/Xayah-Hina]
GitHub: https://github.com/IME-lab-Hokudai/EPINF-NeuFlow
Last Modification Date: 2025-07-10
Description: The pytorch implementation of the paper "Efficient Physics Informed Neural Reconstruction of Diverse Fluid Dynamics from Sparse Observations".
License: Mozilla Public License 2.0 (MPL-2.0)
Copyright (c) 2025 Xiang Yangcheng, IME Lab, Hokkaido University, Japan.
"""
import dataclasses
import math

import tinycudann
import torch

from .api import ModelAPI


class NGPModel(ModelAPI):
    @dataclasses.dataclass
    class Config:
        num_layers_sigma: int = dataclasses.field(default=2, metadata={'help': 'number of layers for the sigma network'})
        hidden_dim_sigma: int = dataclasses.field(default=64, metadata={'help': 'hidden dimension for the sigma network'})
        num_layers_rgb: int = dataclasses.field(default=3, metadata={'help': 'number of layers for the RGB network'})
        hidden_dim_rgb: int = dataclasses.field(default=64, metadata={'help': 'hidden dimension for the RGB network'})
        geo_feat_dim: int = dataclasses.field(default=15, metadata={'help': 'geometric feature dimension'})
        use_background: bool = dataclasses.field(default=False, metadata={'help': 'whether to use background in the dataset'})

    def __init__(self, cfg: Config):
        super().__init__(cfg=cfg)

        bound = 1
        self.encoder_sigma = tinycudann.Encoding(
            n_input_dims=3,
            encoding_config={
                "otype": "HashGrid",
                "n_levels": 16,
                "n_features_per_level": 2,
                "log2_hashmap_size": 19,
                "base_resolution": 16,
                "per_level_scale": 2 ** (math.log2(2048 * bound / 16) / (16 - 1)),
            },
        )
        self.sigma_net = tinycudann.Network(
            n_input_dims=int(self.encoder_sigma.n_output_dims),
            n_output_dims=1 + cfg.geo_feat_dim,
            network_config={
                "otype": "FullyFusedMLP",
                "activation": "ReLU",
                "output_activation": "None",
                "n_neurons": cfg.hidden_dim_sigma,
                "n_hidden_layers": cfg.num_layers_sigma - 1,
            },
        )

        self.encoder_dir = tinycudann.Encoding(
            n_input_dims=3,
            encoding_config={
                "otype": "SphericalHarmonics",
                "degree": 4,
            },
        )

        self.color_net = tinycudann.Network(
            n_input_dims=int(self.encoder_dir.n_output_dims) + cfg.geo_feat_dim,
            n_output_dims=3,
            network_config={
                "otype": "FullyFusedMLP",
                "activation": "ReLU",
                "output_activation": "None",
                "n_neurons": cfg.hidden_dim_rgb,
                "n_hidden_layers": cfg.num_layers_rgb - 1,
            },
        )

    def forward(self, xyz, view_dirs):
        sigma, geo_feat = self.sigma(xyz)
        rgb = self.rgb(xyz, view_dirs, geo_feat)
        return sigma, rgb

    def sigma(self, xyz):
        xyz_encoded = self.encoder_sigma(xyz)
        h = self.sigma_net(xyz_encoded)
        sigma = torch.nn.functional.relu(h[..., :1])
        geo_feat = h[..., 1:]
        return sigma, geo_feat

    def rgb(self, xyz, view_dirs, geo_feat, cond=None):
        view_dirs_encoded = self.encoder_dir(view_dirs)
        h = self.color_net(torch.cat([view_dirs_encoded, geo_feat], dim=-1))
        rgb = torch.sigmoid(h)
        return rgb

    def background(self, sph, view_dirs):
        raise NotImplementedError("Background calculation is not implemented in NGPModel. This model does not support background rendering.")



# -*- coding: utf-8 -*-
"""
Author: Xiang Yangcheng [https://github.com/Xayah-Hina]
GitHub: https://github.com/IME-lab-Hokudai/EPINF-NeuFlow
Last Modification Date: 2025-07-10
Description: The pytorch implementation of the paper "Efficient Physics Informed Neural Reconstruction of Diverse Fluid Dynamics from Sparse Observations".
License: Mozilla Public License 2.0 (MPL-2.0)
Copyright (c) 2025 Xiang Yangcheng, IME Lab, Hokkaido University, Japan.
"""
import dataclasses
import typing

import torch

from .api import RendererAPI
from ..models import EPINFHybridModel
from ..plugins import TrainSpace, SparseLoss, ImageMaskLoss, HyFluidVelocityLoss, RGBAccConsistLoss


class EPINFRenderer(RendererAPI):
    @dataclasses.dataclass
    class Config:
        num_sample_rays: int = dataclasses.field(default=1024 * 2, metadata={'help': 'number of rays to sample per batch'})
        num_depth_samples: int = dataclasses.field(default=1024, metadata={'help': 'number of depth samples per ray'})
        num_importance_samples: int = dataclasses.field(default=0, metadata={'help': 'number of importance samples per ray'})
        use_normalized_directions: bool = dataclasses.field(default=True, metadata={'help': 'whether to use normalized directions for rendering'})
        chunk: int = dataclasses.field(default=1024 * 2, metadata={'help': 'chunk size for rendering'})

        normalized_directions: bool = dataclasses.field(default=True, metadata={'help': 'whether to use normalized directions for rendering'})
        importance_samples: bool = dataclasses.field(default=True, metadata={'help': 'whether to use importance sampling for rendering'})

    def __init__(self, cfg: Config):
        super().__init__(cfg=cfg)

        # ---------- PLUGINS ----------
        self.tsm: TrainSpace = TrainSpace(resolution=64)
        self.register_plugin(plugin=self.tsm)
        self.register_plugin(plugin=SparseLoss(max_level=3, delay_iters=300))
        self.register_plugin(plugin=ImageMaskLoss(delay_iters=300))
        # self.register_plugin(plugin=HyFluidVelocityLoss())
        # self.register_plugin(plugin=RGBAccConsistLoss(delay_iters=300))
        # ---------- PLUGINS ----------

        self.mode: typing.Literal['hybrid', 'dynamic', 'static'] = 'hybrid'
        self.set_mode(mode=self.mode)

    def set_mode(self, mode: typing.Literal['hybrid', 'dynamic', 'static']):
        if mode not in ['hybrid', 'dynamic', 'static', 'vel']:
            raise ValueError(f"Unknown mode: {mode}")
        self.mode = mode

    @property
    def num_sample_rays(self):
        return self.cfg.num_sample_rays

    @property
    def normalized_directions(self):
        return self.cfg.normalized_directions

    @property
    def chunk(self):
        return self.cfg.chunk

    def render_impl(self,
                    model: EPINFHybridModel,
                    rays_o: torch.Tensor,
                    rays_d: torch.Tensor,
                    nears: torch.Tensor,
                    fars: torch.Tensor,
                    pose_indices: torch.Tensor,
                    ):
        match self.mode:
            case 'hybrid':
                return self.render_hybrid(model=model, rays_o=rays_o, rays_d=rays_d, nears=nears, fars=fars, pose_indices=pose_indices)
            case 'dynamic':
                return self.render_dynamic(model=model, rays_o=rays_o, rays_d=rays_d, nears=nears, fars=fars, pose_indices=pose_indices)
            case 'static':
                return self.render_static(model=model, rays_o=rays_o, rays_d=rays_d, nears=nears, fars=fars, pose_indices=pose_indices)
            case _:
                raise ValueError(f"Unknown mode: {self.mode}")

    def render_hybrid(self,
                      model: EPINFHybridModel,
                      rays_o: torch.Tensor,
                      rays_d: torch.Tensor,
                      nears: torch.Tensor,
                      fars: torch.Tensor,
                      pose_indices: torch.Tensor,
                      ):
        num_sample_rays = rays_d.shape[0]

        xyz, z_vals = self.sample_points(
            rays_o=rays_o,
            rays_d=rays_d,
            nears=nears,
            fars=fars,
            num_depth_samples=self.cfg.num_depth_samples,
            use_perturb=model.training,
        )
        xyz_masked, mask = self.normalize_and_filter_xyz(xyz=xyz)
        view_dirs_masked = self.normalize_and_filter_view_dirs(view_dirs=rays_d, expand_shape=xyz.shape, mask=mask)

        if xyz_masked.numel() > 0:
            sigma_static_masked, geo_feat_static_masked = model.static_renderer.sigma(xyz_masked)
            sigma_dynamic_masked, geo_feat_dynamic_masked = model.dynamic_renderer.sigma(xyz_masked)
            sigma_static = self.assemble_masked_tensor(mask, sigma_static_masked, (num_sample_rays, self.cfg.num_depth_samples, 1))  # shape: (num_sample_rays, num_depth_samples, 1)
            sigma_dynamic = self.assemble_masked_tensor(mask, sigma_dynamic_masked, (num_sample_rays, self.cfg.num_depth_samples, 1))  # shape: (num_sample_rays, num_depth_samples, 1)

            rgb_static_masked = model.static_renderer.rgb(None, view_dirs_masked, geo_feat_static_masked)
            rgb_dynamic_masked = model.dynamic_renderer.rgb(None, view_dirs_masked, geo_feat_dynamic_masked)
            rgb_static = self.assemble_masked_tensor(mask, rgb_static_masked, (num_sample_rays, self.cfg.num_depth_samples, 3))  # shape: (num_sample_rays, num_depth_samples, 3)
            rgb_dynamic = self.assemble_masked_tensor(mask, rgb_dynamic_masked, (num_sample_rays, self.cfg.num_depth_samples, 3))  # shape: (num_sample_rays, num_depth_samples, 3)

            alpha_static = self.sigma_to_alpha(sigma=sigma_static, z_vals=z_vals)  # shape: (num_sample_rays, num_depth_samples)
            alpha_dynamic = self.sigma_to_alpha(sigma=sigma_dynamic, z_vals=z_vals)  # shape: (num_sample_rays, num_depth_samples)
            alpha_hybrid = 1 - (1.0 - alpha_static) * (1.0 - alpha_dynamic)  # shape: (num_sample_rays, num_depth_samples)

            eps = torch.finfo(alpha_hybrid.dtype).eps
            transmittance_hybrid = torch.cumprod(
                torch.cat([torch.ones_like(alpha_hybrid[:, :1]), 1. - alpha_hybrid + eps], dim=-1),
                dim=-1,
            )[:, :-1]  # shape: (num_sample_rays, num_depth_samples)
            weights_static = alpha_static * transmittance_hybrid  # shape: (num_sample_rays, num_depth_samples)
            weights_dynamic = alpha_dynamic * transmittance_hybrid  # shape: (num_sample_rays, num_depth_samples)

            acc_map = self.weights_to_acc_map(weights=weights_static) + self.weights_to_acc_map(weights=weights_dynamic)  # shape: (num_sample_rays,)
            depth_map = self.weights_to_depth_map(weights=weights_static, z_vals=z_vals, nears=nears, fars=fars) + self.weights_to_depth_map(weights=weights_dynamic, z_vals=z_vals, nears=nears, fars=fars)  # shape: (num_sample_rays,)
            rgb_map = self.weights_to_rgb_map(weights=weights_static, rgb=rgb_static) + self.weights_to_rgb_map(weights=weights_dynamic, rgb=rgb_dynamic)  # shape: (num_sample_rays, 3)
            rgb_map = rgb_map + (1 - acc_map).unsqueeze(-1) * self.background_color

            # ---------- PLUGINS ----------
            if self.training:
                self.tsm.update(xyz=xyz_masked, index=int(pose_indices[0].item()))
                for sl in filter(lambda v: isinstance(v, SparseLoss), self.plugins):
                    sl: SparseLoss
                    sl.compute(tsm=self.tsm, xyz=xyz_masked, sigma=sigma_static_masked + sigma_dynamic_masked)
            # ---------- PLUGINS ----------

        else:
            rgb_map = torch.zeros_like(rays_d)  # shape: (num_sample_rays, 3)
            depth_map = torch.zeros_like(rays_d[..., 0])  # shape: (num_sample_rays,)
            acc_map = torch.zeros_like(rays_d[..., 0])  # shape: (num_sample_rays,)

        ret_static = self.render_static(model, rays_o=rays_o, rays_d=rays_d, nears=nears, fars=fars, pose_indices=pose_indices)
        ret_dynamic = self.render_dynamic(model, rays_o=rays_o, rays_d=rays_d, nears=nears, fars=fars, pose_indices=pose_indices)

        return {
            "rgb_map": rgb_map,
            "depth_map": depth_map,
            "acc_map": acc_map,
            'rgb_map_static_independent': ret_static['rgb_map'],
            'depth_map_static_independent': ret_static['depth_map'],
            'acc_map_static_independent': ret_static['acc_map'],
            'rgb_map_dynamic_independent': ret_dynamic['rgb_map'],
            'depth_map_dynamic_independent': ret_dynamic['depth_map'],
            'acc_map_dynamic_independent': ret_dynamic['acc_map'],
        }

    def render_dynamic(self,
                       model: EPINFHybridModel,
                       rays_o: torch.Tensor,
                       rays_d: torch.Tensor,
                       nears: torch.Tensor,
                       fars: torch.Tensor,
                       pose_indices: torch.Tensor,
                       ):
        num_sample_rays = rays_d.shape[0]

        xyz, z_vals = self.sample_points(
            rays_o=rays_o,
            rays_d=rays_d,
            nears=nears,
            fars=fars,
            num_depth_samples=self.cfg.num_depth_samples,
            use_perturb=model.training,
        )
        xyz_masked, mask = self.normalize_and_filter_xyz(xyz=xyz)
        view_dirs_masked = self.normalize_and_filter_view_dirs(view_dirs=rays_d, expand_shape=xyz.shape, mask=mask)
        if xyz_masked.numel() > 0:
            sigma_ret = model.dynamic_renderer.sigma(xyz_masked)
            sigma_dynamic_masked, geo_feat_dynamic_masked = sigma_ret[0], sigma_ret[1]
            sigma_dynamic = self.assemble_masked_tensor(mask, sigma_dynamic_masked, (num_sample_rays, self.cfg.num_depth_samples, 1))  # shape: (num_sample_rays, num_depth_samples, 1)
            rgb_dynamic_masked = model.dynamic_renderer.rgb(None, view_dirs_masked, geo_feat_dynamic_masked)
            rgb_dynamic = self.assemble_masked_tensor(mask, rgb_dynamic_masked, (num_sample_rays, self.cfg.num_depth_samples, 3))  # shape: (num_sample_rays, num_depth_samples, 3)
            alpha_dynamic = self.sigma_to_alpha(sigma=sigma_dynamic, z_vals=z_vals)  # shape: (num_sample_rays, num_depth_samples)
            weights_dynamic_independent = self.alpha_to_weights(alpha=alpha_dynamic)
            acc_map_dynamic_independent = self.weights_to_acc_map(weights=weights_dynamic_independent)
            depth_map_dynamic_independent = self.weights_to_depth_map(weights=weights_dynamic_independent, z_vals=z_vals, nears=nears, fars=fars)
            rgb_map_dynamic_independent = self.weights_to_rgb_map(weights=weights_dynamic_independent, rgb=rgb_dynamic)
            rgb_map_dynamic_independent = rgb_map_dynamic_independent + (1 - acc_map_dynamic_independent).unsqueeze(-1) * self.background_color

            if self.training:
                for hvl in filter(lambda v: isinstance(v, HyFluidVelocityLoss), self.plugins):
                    hvl: HyFluidVelocityLoss
                    assert len(sigma_ret) == 4, "Expected sigma_ret to have 4 elements: (sigma_dynamic_masked, geo_feat_dynamic_masked, xyzt_encoded, xyzt)"
                    xyzt_encoded, xyzt = sigma_ret[2], sigma_ret[3]
                    hvl.compute(
                        xyzt=xyzt,
                        xyzt_encoded=xyzt_encoded,
                        sigma=sigma_dynamic_masked,
                        vel=model.vel_renderer.sigma_grad(xyzt_grad=xyzt),
                        fn=model.fn,
                    )
        else:
            rgb_map_dynamic_independent = torch.zeros_like(rays_d)  # shape: (num_sample_rays, 3)
            depth_map_dynamic_independent = torch.zeros_like(rays_d[..., 0])  # shape: (num_sample_rays,)
            acc_map_dynamic_independent = torch.zeros_like(rays_d[..., 0])  # shape: (num_sample_rays,)

        return {
            "rgb_map": rgb_map_dynamic_independent,
            "depth_map": depth_map_dynamic_independent,
            "acc_map": acc_map_dynamic_independent,
        }

    def render_static(self,
                      model: EPINFHybridModel,
                      rays_o: torch.Tensor,
                      rays_d: torch.Tensor,
                      nears: torch.Tensor,
                      fars: torch.Tensor,
                      pose_indices: torch.Tensor,
                      ):
        num_sample_rays = rays_d.shape[0]

        xyz, z_vals = self.sample_points(
            rays_o=rays_o,
            rays_d=rays_d,
            nears=nears,
            fars=fars,
            num_depth_samples=self.cfg.num_depth_samples,
            use_perturb=model.training,
        )
        xyz_masked, mask = self.normalize_and_filter_xyz(xyz=xyz)
        view_dirs_masked = self.normalize_and_filter_view_dirs(view_dirs=rays_d, expand_shape=xyz.shape, mask=mask)
        if xyz_masked.numel() > 0:
            sigma_static_masked, geo_feat_static_masked = model.static_renderer.sigma(xyz_masked)
            sigma_static = self.assemble_masked_tensor(mask, sigma_static_masked, (num_sample_rays, self.cfg.num_depth_samples, 1))  # shape: (num_sample_rays, num_depth_samples, 1)
            rgb_static_masked = model.static_renderer.rgb(None, view_dirs_masked, geo_feat_static_masked)
            rgb_static = self.assemble_masked_tensor(mask, rgb_static_masked, (num_sample_rays, self.cfg.num_depth_samples, 3))  # shape: (num_sample_rays, num_depth_samples, 3)
            alpha_static = self.sigma_to_alpha(sigma=sigma_static, z_vals=z_vals)  # shape: (num_sample_rays, num_depth_samples)
            weights_static_independent = self.alpha_to_weights(alpha=alpha_static)
            acc_map_static_independent = self.weights_to_acc_map(weights=weights_static_independent)
            depth_map_static_independent = self.weights_to_depth_map(weights=weights_static_independent, z_vals=z_vals, nears=nears, fars=fars)
            rgb_map_static_independent = self.weights_to_rgb_map(weights=weights_static_independent, rgb=rgb_static)
            rgb_map_static_independent = rgb_map_static_independent + (1 - acc_map_static_independent).unsqueeze(-1) * self.background_color

        else:
            rgb_map_static_independent = torch.zeros_like(rays_d)  # shape: (num_sample_rays, 3)
            depth_map_static_independent = torch.zeros_like(rays_d[..., 0])  # shape: (num_sample_rays,)
            acc_map_static_independent = torch.zeros_like(rays_d[..., 0])  # shape: (num_sample_rays,)

        return {
            "rgb_map": rgb_map_static_independent,
            "depth_map": depth_map_static_independent,
            "acc_map": acc_map_static_independent,
        }



# -*- coding: utf-8 -*-
"""
Author: Xiang Yangcheng [https://github.com/Xayah-Hina]
GitHub: https://github.com/IME-lab-Hokudai/EPINF-NeuFlow
Last Modification Date: 2025-07-10
Description: The pytorch implementation of the paper "Efficient Physics Informed Neural Reconstruction of Diverse Fluid Dynamics from Sparse Observations".
License: Mozilla Public License 2.0 (MPL-2.0)
Copyright (c) 2025 Xiang Yangcheng, IME Lab, Hokkaido University, Japan.
"""
from .api import PIDatasetAPI
from .nerf_synthetic import NeRFSyntheticDataset
from .pi_neuflow import PINeuFlowDataset

__all__ = ['PIDatasetAPI', 'NeRFSyntheticDataset', 'PINeuFlowDataset']



# -*- coding: utf-8 -*-
"""
Author: Xiang Yangcheng [https://github.com/Xayah-Hina]
GitHub: https://github.com/IME-lab-Hokudai/EPINF-NeuFlow
Last Modification Date: 2025-07-12
Description: The pytorch implementation of the paper "Efficient Physics Informed Neural Reconstruction of Diverse Fluid Dynamics from Sparse Observations".
License: Mozilla Public License 2.0 (MPL-2.0)
Copyright (c) 2025 Xiang Yangcheng, IME Lab, Hokkaido University, Japan.
"""
import torch

from .api import PluginLoss
from .train_sapce import TrainSpace


class ImageMaskLoss(PluginLoss):
    def __init__(self, delay_iters: int = 0, name: str = 'image_mask_loss'):
        super().__init__(name=name, weight=10000)
        self.delay_iters = delay_iters
        self.total_iters = 0

    def compute(self, rgb_map, acc_map, pixels_mask, background_color):
        self.total_iters += 1
        if self.total_iters < self.delay_iters:
            return

        masked_rgb_map = rgb_map[~pixels_mask]
        if masked_rgb_map.numel() > 0:
            masked_rgb_map = masked_rgb_map.view(-1, 3)
            background_color = background_color.view(1, 3)
            self._loss += torch.mean(torch.sum((masked_rgb_map - background_color) ** 2, dim=-1))
        masked_acc_map = acc_map[~pixels_mask]
        if masked_acc_map.numel() > 0:
            masked_acc_map = masked_acc_map.view(-1)
            self._loss += torch.mean(masked_acc_map)


class SparseLoss(PluginLoss):
    def __init__(self, max_level: int, delay_iters: int, name: str = 'sparse_loss'):
        super().__init__(name=name, weight=10000)
        assert max_level < 4, "max_level should be less than 4, as the maximum level in EPINF is 3."

        self.max_level = max_level
        self.delay_iters = delay_iters
        self.total_iters = 0

    def compute(self, tsm: TrainSpace, xyz: torch.Tensor, sigma: torch.Tensor):
        self.total_iters += 1
        if self.total_iters < self.delay_iters:
            return

        for level in range(self.max_level):
            sparse_mask = tsm.exclusive_n_mask(xyz=xyz, n=level + 1)
            sigma_sparse = sigma[sparse_mask]
            if sigma_sparse.numel() > 0:
                sparse_loss = 10.0 ** (-level) * sigma_sparse.mean()
                self._loss += sparse_loss


class RGBAccConsistLoss(PluginLoss):
    def __init__(self, delay_iters: int = 0, name: str = 'rgb_acc_consist_loss'):
        super().__init__(name=name, weight=10000)
        self.delay_iters = delay_iters
        self.total_iters = 0

    def compute(self, rgb_map, acc_map):
        self.total_iters += 1
        if self.total_iters < self.delay_iters:
            return

        if rgb_map.numel() > 0 and acc_map.numel() > 0:
            rgb_map_gray = rgb_map.mean(dim=-1, keepdim=True)
            rgb_map_gray_norm = rgb_map_gray / (rgb_map_gray.max() + 1e-6)
            acc_map_norm = acc_map / (acc_map.max() + 1e-6)
            proportional_loss = torch.mean((rgb_map_gray_norm - acc_map_norm) ** 2)
            grayscale_loss = torch.mean((rgb_map - rgb_map_gray) ** 2)
            self._loss += (proportional_loss + 0.1 * grayscale_loss)


class HyFluidVelocityLoss(PluginLoss):
    def __init__(self, delay_iters: int = 0, name: str = "hyfluid_velocity_loss"):
        super().__init__(name=name, weight=1)
        self.total_iters = 0
        self.delay_iters = delay_iters

    @staticmethod
    def get_minibatch_jacobian(y, x):
        """Computes the Jacobian of y wrt x assuming minibatch-mode.
        Args:
          y: (N, ...) with a total of D_y elements in ...
          x: (N, ...) with a total of D_x elements in ...
        Returns:
          The minibatch Jacobian matrix of shape (N, D_y, D_x)
        """
        assert y.shape[0] == x.shape[0]
        y = y.view(y.shape[0], -1)
        # Compute Jacobian row by row.
        jac = []
        for j in range(y.shape[1]):
            dy_j_dx = torch.autograd.grad(
                y[:, j],
                x,
                torch.ones_like(y[:, j], device=y.get_device()),
                retain_graph=True,
                create_graph=True,
            )[0].view(x.shape[0], -1)

            jac.append(torch.unsqueeze(dy_j_dx, 1))
        jac = torch.cat(jac, 1)
        return jac

    def compute(self, xyzt, xyzt_encoded, sigma, vel, fn):
        self.total_iters += 1

        if self.total_iters < self.delay_iters:
            return

        split_nse_wei = 0.001

        jac = torch.vmap(torch.func.jacrev(fn))(xyzt_encoded)
        jac = jac.contiguous()
        jac_x = self.get_minibatch_jacobian(xyzt_encoded, xyzt)
        jac_x = jac_x.contiguous()
        jac = jac @ jac_x

        _d_x, _d_y, _d_z, _d_t = [torch.squeeze(_, -1) for _ in jac.split(1, dim=-1)]

        _u_x, _u_y, _u_z, _u_t = None, None, None, None
        _u, _v, _w = vel.split(1, dim=-1)
        split_nse = _d_t + (_u * _d_x + _v * _d_y + _w * _d_z)
        nse_errors = torch.mean(torch.square(split_nse))

        viz_dens_mask = sigma.detach() > 0.1
        vel_norm = vel.norm(dim=-1, keepdim=True)
        min_vel_mask = vel_norm.detach() < 0.2 * sigma.detach()
        vel_reg_mask = min_vel_mask & viz_dens_mask
        min_vel_reg_map = (0.2 * sigma - vel_norm) * vel_reg_mask.float()
        min_vel_reg = min_vel_reg_map.pow(2).mean()

        zero_dens_mask = sigma.detach() < 1e-3
        keep_zero_map = vel_norm * zero_dens_mask.float()
        proj_loss = split_nse_wei * keep_zero_map.pow(2).mean()

        self._loss += nse_errors + 10 * min_vel_reg + proj_loss



# -*- coding: utf-8 -*-
"""
Author: Xiang Yangcheng [https://github.com/Xayah-Hina]
GitHub: https://github.com/IME-lab-Hokudai/EPINF-NeuFlow
Last Modification Date: 2025-07-10
Description: The pytorch implementation of the paper "Efficient Physics Informed Neural Reconstruction of Diverse Fluid Dynamics from Sparse Observations".
License: Mozilla Public License 2.0 (MPL-2.0)
Copyright (c) 2025 Xiang Yangcheng, IME Lab, Hokkaido University, Japan.
"""
import dataclasses
import json
import os
import typing

import rich.progress
import torch
import torchvision

from .api import PIDataset, PIDatasetAPI


class NeRFSynthetic(PIDataset):
    """
    PRIVATE CLASS
    - This class and its subclasses are only used inside datasets package.
    """

    def __init__(self, dataset_path: str, dataset_type: typing.Literal['train', 'val', 'test'], downscale: int, use_fp16: bool, use_image_mask: bool):
        super().__init__()
        with open(os.path.join(dataset_path, 'transforms_' + dataset_type + '.json'), 'r') as f:
            transform = json.load(f)
            images = []
            poses = []
            with rich.progress.Progress() as progress:
                task = progress.add_task(f"[cyan]Loading Images ({dataset_type})...", total=len(transform['frames']))
                for frame in transform['frames']:
                    frame_path = os.path.join(dataset_path, frame['file_path'] + '.png')
                    image = torchvision.io.read_image(path=str(frame_path)).permute(1, 2, 0) / 255.0  # (H, W, C)
                    images.append(image)
                    pose = torch.tensor(frame['transform_matrix'])
                    poses.append(pose)
                    progress.update(task, advance=1)
            images = torch.stack(images, dim=0)
            poses = torch.stack(poses, dim=0)
            N, H, W, C = images.shape
            if downscale != 1:
                H, W = H // downscale, W // downscale
                images = torch.nn.functional.interpolate(images.permute(0, 3, 1, 2), size=(H, W), mode='bilinear', align_corners=False).permute(0, 2, 3, 1)
            focal = W / (2 * torch.tan(torch.tensor(transform['camera_angle_x']) / 2))

        self._images = images.to(dtype=torch.float16 if use_fp16 else torch.float32)
        self._poses = poses.to(dtype=torch.float16 if use_fp16 else torch.float32)
        self._focal = focal.to(dtype=torch.float16 if use_fp16 else torch.float32)

        self._near = 0.1
        self._far = 5.0

        self._use_image_mask = use_image_mask

    @property
    def images(self):
        return self._images

    @property
    def images_highlighted(self):
        return None

    @property
    def images_masks(self):
        return self.images[..., 3] > 0

    @property
    def poses(self):
        return self._poses

    @property
    def focal(self):
        return self._focal

    @property
    def width(self):
        return self.images.shape[-2]

    @property
    def height(self):
        return self.images.shape[-3]

    @property
    def near(self):
        return self._near

    @property
    def far(self):
        return self._far

    @property
    def num_poses(self):
        return self.poses.shape[0]

    @property
    def num_times(self):
        return 1

    def transform(self, translate, scale):
        self._poses[:, :3, 3] = (self._poses[:, :3, 3] + torch.as_tensor(translate, dtype=self._poses.dtype, device=self._poses.device).view(1, 3)) * scale
        self._near = self._near * scale
        self._far = self._far * scale

    def __getitem__(self, index):
        return {
            'image': self.images[index],
            'pose': self.poses[index],
            'focal': self.focal,
            'index': index,
            'min_near': self.near,
            'max_far': self.far,
        }

    def __len__(self):
        return len(self.images)


class NeRFSyntheticDataset(PIDatasetAPI):
    """
    BATCH DATA FORMAT:
        images: (N, H, W, C)
        images_masks: (N, H, W)
        poses: (N, 4, 4)
        focals: (N)
        pose_indices: (N)
        min_nears: (N)
        max_fars: (N)
    PUBLIC API:
        dataset_path: str
        aabb_std: list[min_x, min_y, min_z, max_x, max_y, max_z]
        bound_std: list[min_x, min_y, min_z, max_x, max_y, max_z]
        background_color: list[r, g, b]
        complex_background: bool
        num_train_poses: int
    """

    @dataclasses.dataclass
    class Config:
        base_data_dir: str = dataclasses.field(default='data/nerf_synthetic/', metadata={'help': 'base data directory'})
        dataset: typing.Literal['chair', 'drums', 'ficus', 'hotdog', 'lego', 'materials', 'mic', 'ship'] = dataclasses.field(default='lego', metadata={'help': 'dataset name'})
        downscale: int = dataclasses.field(default=1, metadata={'help': 'downscale factor for the dataset images'})
        downscale_val: int = dataclasses.field(default=2, metadata={'help': 'downscale factor for the validation images'})
        fp16: bool = dataclasses.field(default=True, metadata={'help': 'use mixed precision training'})

        pre_translate: list[float] = dataclasses.field(default_factory=lambda: [0.0, 0.0, 0.0], metadata={'help': 'translation to apply to the dataset poses'})
        pre_scale: float = dataclasses.field(default=0.7, metadata={'help': 'scale to apply to the dataset poses'})
        aabb_std: list[float] = dataclasses.field(default_factory=lambda: [-1.0, -1.0, -1.0, 1.0, 1.0, 1.0], metadata={'help': 'standard axis-aligned bounding box for the dataset'})
        bound_std: list[float] = dataclasses.field(default_factory=lambda: [-1.0, -1.0, -1.0, 1.0, 1.0, 1.0], metadata={'help': 'standard bounding box for the dataset'})

    def __init__(self, cfg: Config):
        super().__init__(cfg=cfg)

    def prepare_data(self):
        from huggingface_hub import snapshot_download
        snapshot_download(
            repo_id="XayahHina/nerf_synthetic",
            repo_type="dataset",
            local_dir=self.cfg.base_data_dir,
        )
        if not os.path.exists(os.path.join(self.cfg.base_data_dir, "README.txt")):
            import zipfile
            with zipfile.ZipFile(os.path.join(self.cfg.base_data_dir, "nerf_synthetic.zip"), 'r') as zip_ref:
                zip_ref.extractall(self.cfg.base_data_dir)

    def setup(self, stage: str):
        if "lego" in (self.cfg.base_data_dir + self.cfg.dataset).lower():
            self.cfg.bound_std = [-0.68, -1.0, -0.6, 0.68, 1.0, 1.0]
        super().setup(stage)

    def dataset(self, dataset_type: typing.Literal['train', 'val', 'test']) -> PIDataset:
        downscale = self.cfg.downscale_val if dataset_type == 'val' else self.cfg.downscale
        dataset = NeRFSynthetic(dataset_path=self.cfg.base_data_dir + self.cfg.dataset, dataset_type=dataset_type, downscale=downscale, use_fp16=self.cfg.fp16, use_image_mask=self.complex_background)
        dataset.transform(translate=self.pre_translate, scale=self.pre_scale)
        return dataset

    @staticmethod
    def collate(batch: list):
        images = torch.stack([single['image'] for single in batch])
        poses = torch.stack([single['pose'] for single in batch])
        focals = torch.tensor([single['focal'] for single in batch])
        pose_indices = torch.tensor([single['index'] for single in batch])
        min_nears = torch.tensor([single['min_near'] for single in batch]).to(poses.dtype)
        max_fars = torch.tensor([single['max_far'] for single in batch]).to(poses.dtype)

        return {
            'images': images,
            'images_masks': images[..., 3] > 0,
            'poses': poses,
            'focals': focals,
            'pose_indices': pose_indices,
            'min_nears': min_nears,
            'max_fars': max_fars,
        }

    @property
    def pre_translate(self):
        return self.cfg.pre_translate

    @property
    def pre_scale(self):
        return self.cfg.pre_scale

    @property
    def aabb_std(self):
        return self.cfg.aabb_std

    @property
    def bound_std(self):
        return self.cfg.bound_std

    @property
    def background_color(self):
        return [0.0, 0.0, 0.0]  # Hardcoded for NeRFSynthetic, as it does not use complex backgrounds

    @property
    def complex_background(self):
        return False  # Hardcoded for NeRFSynthetic, as it does not use complex backgrounds



# -*- coding: utf-8 -*-
"""
Author: Xiang Yangcheng [https://github.com/Xayah-Hina]
GitHub: https://github.com/IME-lab-Hokudai/EPINF-NeuFlow
Last Modification Date: 2025-07-10
Description: The pytorch implementation of the paper "Efficient Physics Informed Neural Reconstruction of Diverse Fluid Dynamics from Sparse Observations".
License: Mozilla Public License 2.0 (MPL-2.0)
Copyright (c) 2025 Xiang Yangcheng, IME Lab, Hokkaido University, Japan.
"""
import torch


class ModelAPI(torch.nn.Module):
    def __init__(self, cfg):
        super().__init__()
        self.cfg = cfg

    def sigma(self, xyz):
        raise NotImplementedError("Sigma calculation is not implemented in the base API class. Please implement this method in a subclass.")

    def rgb(self, xyz, view_dirs, geo_feat, cond=None):
        raise NotImplementedError("RGB calculation is not implemented in the base API class. Please implement this method in a subclass.")

    def background(self, sph, view_dirs):
        raise NotImplementedError("Background calculation is not implemented in the base API class. Please implement this method in a subclass.")



# -*- coding: utf-8 -*-
"""
Author: Xiang Yangcheng [https://github.com/Xayah-Hina]
GitHub: https://github.com/IME-lab-Hokudai/EPINF-NeuFlow
Last Modification Date: 2025-07-10
Description: The pytorch implementation of the paper "Efficient Physics Informed Neural Reconstruction of Diverse Fluid Dynamics from Sparse Observations".
License: Mozilla Public License 2.0 (MPL-2.0)
Copyright (c) 2025 Xiang Yangcheng, IME Lab, Hokkaido University, Japan.
"""
import dataclasses
import glob
import multiprocessing
import os
import subprocess
import tempfile
import textwrap

import lightning
import numpy as np
import rich
import rich.progress
import torch
import torchmetrics
import torchvision.utils

from .datasets import PIDatasetAPI, PINeuFlowDataset
from .fluids import advect_density
from .models import EPINFHybridModel
from .plugins import PluginLoss
from .renderers import EPINFRenderer


class EPINFTrainer(lightning.LightningModule):
    @dataclasses.dataclass
    class EPINF:
        dataset: PINeuFlowDataset.Config = dataclasses.field(default_factory=PINeuFlowDataset.Config, metadata={'help': 'PI-NeuFlow dataset configuration'})
        model: EPINFHybridModel.Config = dataclasses.field(default_factory=EPINFHybridModel.Config, metadata={'help': 'model configuration'})
        renderer: EPINFRenderer.Config = dataclasses.field(default_factory=EPINFRenderer.Config, metadata={'help': 'renderer configuration'})

        # general settings
        gpu_id: int = dataclasses.field(default=0, metadata={'help': 'GPU ID to use for training'})
        gui: bool = dataclasses.field(default=False, metadata={'help': 'whether to use GUI for rendering'})
        ckpt: bool = dataclasses.field(default=False, metadata={'help': 'checkpoint path to load for testing'})

        # training settings
        epochs: int = dataclasses.field(default=100, metadata={'help': 'number of epochs to train the model'})
        lrate: float = dataclasses.field(default=1e-3, metadata={'help': 'learning rate for the optimizer'})
        val_skip: int = dataclasses.field(default=10, metadata={'help': 'number of epochs to skip for validation'})
        save_val_map: bool = dataclasses.field(default=True, metadata={'help': 'whether to output validation images'})
        export_model_only: bool = dataclasses.field(default=False, metadata={'help': 'whether to export model only'})

        # test settings
        test: bool = dataclasses.field(default=False, metadata={'help': 'test mode, load a checkpoint and test the model'})
        export: bool = dataclasses.field(default=False, metadata={'help': 'whether to export fields'})
        export_houdini: bool = dataclasses.field(default=False, metadata={'help': 'whether to export fields for Houdini'})
        fields: bool = dataclasses.field(default=True, metadata={'help': 'whether to export fields, defaults to True when export or export_houdini is True'})
        resim: bool = dataclasses.field(default=False, metadata={'help': 'whether to resimulate the fields'})

        fading_step: int = dataclasses.field(default=3000, metadata={'help': 'step at which the static fading starts'})

    def __init__(self, cfg: EPINF):
        super().__init__()
        self.cfg = cfg
        self.save_hyperparameters()
        self.model = EPINFHybridModel(cfg=cfg.model)
        self.renderer = EPINFRenderer(cfg=cfg.renderer)
        self.psnr = torchmetrics.image.PeakSignalNoiseRatio(data_range=1.0)
        self.ssim = torchmetrics.image.StructuralSimilarityIndexMeasure(data_range=1.0)
        self.lpips = torchmetrics.image.lpip.LearnedPerceptualImagePatchSimilarity(net_type='vgg')

    def on_fit_start(self):
        dataset: PIDatasetAPI = self.trainer.datamodule
        self.renderer.set_render_resolution(width=dataset.width, height=dataset.height)
        rich.print(f"[bold green]>>> [Custom]: Renderer resolution set to {dataset.width}x{dataset.height} <<<[/bold green]")
        self.renderer.set_min_near_far(min_near=dataset.near_std, max_far=dataset.far_std)
        rich.print(f"[bold green]>>> [Custom]: Near far std set to {dataset.near_std} - {dataset.far_std} <<<[/bold green]")
        self.renderer.set_aabb_std(dataset.aabb_std)
        rich.print(f"[bold green]>>> [Custom]: AABB std set to {dataset.aabb_std} <<<[/bold green]")
        self.renderer.set_bound_std(dataset.bound_std)
        rich.print(f"[bold green]>>> [Custom]: Bound std set to {dataset.bound_std} <<<[/bold green]")
        self.renderer.set_background_color(dataset.background_color)
        rich.print(f"[bold green]>>> [Custom]: Background color set to {dataset.background_color} <<<[/bold green]")
        self.renderer.to(device=self.device)
        rich.print(f"[bold green]>>> [Custom]: Renderer moved to device {self.device} <<<[/bold green]")

    def on_train_batch_start(self, batch, batch_idx):
        self.model.set_time(batch['times'])
        for plugin in self.renderer.plugins:
            plugin.on_iter_start()

    def training_step(self, batch, batch_idx):
        result_maps, pixels, pixels_mask = self.renderer.forward(
            model=self.model,
            poses=batch['poses'],
            focals=batch['focals'],
            images=batch['images'],
            pose_indices=batch.get('pose_indices', None),
            images_masks=batch.get('images_masks', None),
        )

        rgb_map = result_maps['rgb_map']
        img_loss = 10000 * torch.nn.functional.mse_loss(rgb_map, pixels[..., :3])
        loss = img_loss
        self.log("img_loss", img_loss, on_step=True, on_epoch=False, prog_bar=True)

        for plugin_loss in filter(lambda v: isinstance(v, PluginLoss), self.renderer.plugins):
            p_loss = plugin_loss.weight * plugin_loss.loss
            loss += p_loss
            self.log(f"{plugin_loss.name}", p_loss, on_step=True, on_epoch=False, prog_bar=True)

        if 'rgb_map_static_independent' in result_maps:
            tempo_fading = min(max(self.global_step / self.cfg.fading_step, 0.0), 1.0)
            img_loss_static = 10000 * torch.nn.functional.mse_loss(result_maps['rgb_map_static_independent'], pixels[..., :3])
            img_static_loss = tempo_fading * img_loss + (1 - tempo_fading) * img_loss_static
            loss += img_static_loss
            self.log("img_static_loss", img_static_loss, on_step=True, on_epoch=False, prog_bar=True)

        self.log("loss", loss, on_step=True, on_epoch=False, prog_bar=True)
        return loss

    def on_validation_batch_start(self, batch, batch_idx, dataloader_idx=0):
        if self.cfg.save_val_map:
            os.makedirs(os.path.join(self.logger.experiment.dir, 'validation_images'), exist_ok=True)
        self.model.set_time(batch['times'])

    def validation_step(self, batch, batch_idx):
        if batch_idx % self.cfg.val_skip != 0:
            return None

        result_maps, pixels, pixels_mask = self.renderer.forward(
            model=self.model,
            poses=batch['poses'],
            focals=batch['focals'],
            images=batch['images'],
            pose_indices=batch.get('pose_indices', None),
            images_masks=batch.get('images_masks', None),
        )

        val_loss = torch.nn.functional.mse_loss(result_maps['rgb_map'], pixels[..., :3])
        self.log("val_loss", val_loss, on_step=False, on_epoch=True, prog_bar=True)
        psnr = self.psnr(result_maps['rgb_map'], pixels[..., :3])
        self.log("val_psnr", psnr, on_step=False, on_epoch=True, prog_bar=True)
        ssim = self.ssim(result_maps['rgb_map'].permute(0, 3, 1, 2), pixels[..., :3].permute(0, 3, 1, 2))  # 如果是 HWC 格式，需转为 NCHW
        self.log("val_ssim", ssim, on_step=False, on_epoch=True, prog_bar=True)
        lpips = self.lpips(result_maps['rgb_map'].permute(0, 3, 1, 2), pixels[..., :3].permute(0, 3, 1, 2))
        self.log("val_lpips", lpips, on_step=False, on_epoch=True, prog_bar=True)

        if self.cfg.save_val_map:
            final_map = self.generate_map(result_maps, pixels, batch.get('images_highlighted', None), pixels_mask)
            torchvision.utils.save_image(final_map.permute(2, 0, 1), os.path.join(self.logger.experiment.dir, 'validation_images', f'{self.current_epoch}_{batch_idx}.png'))

        return val_loss

    def on_test_start(self):
        os.makedirs(os.path.join(self.logger.experiment.dir, 'test_images'), exist_ok=True)

    def test_step(self, batch, batch_idx):
        self.model.set_time(batch['times'])
        result_maps, pixels, pixels_mask = self.renderer.forward(
            model=self.model,
            poses=batch['poses'],
            focals=batch['focals'],
            images=batch.get('images', None),
            pose_indices=batch.get('pose_indices', None),
            images_masks=batch.get('images_masks', None),
        )
        final_map = self.generate_map(result_maps, pixels, batch.get('images_highlighted', None), pixels_mask)
        psnr = self.psnr(result_maps['rgb_map'], pixels[..., :3])
        self.log("test_psnr", psnr, on_step=False, on_epoch=True, prog_bar=True)
        ssim = self.ssim(result_maps['rgb_map'].permute(0, 3, 1, 2), pixels[..., :3].permute(0, 3, 1, 2))  # 如果是 HWC 格式，需转为 NCHW
        self.log("test_ssim", ssim, on_step=False, on_epoch=True, prog_bar=True)
        lpips = self.lpips(result_maps['rgb_map'].permute(0, 3, 1, 2), pixels[..., :3].permute(0, 3, 1, 2))
        self.log("test_lpips", lpips, on_step=False, on_epoch=True, prog_bar=True)
        os.makedirs(os.path.join(self.logger.experiment.dir, 'test_images'), exist_ok=True)
        torchvision.utils.save_image(final_map.permute(2, 0, 1), os.path.join(self.logger.experiment.dir, 'test_images', f'{batch_idx}.png'))

    def on_test_end(self):
        image_dir = os.path.join(self.logger.experiment.dir, 'test_images')
        video_path = os.path.join(self.logger.experiment.dir, 'test_images', 'test_video.mp4')

        image_paths = sorted(
            glob.glob(os.path.join(image_dir, '*.png')),
            key=lambda x: int(os.path.basename(x).split('_')[-1].split('.')[0])
        )

        frames = []
        for path in image_paths:
            img = torchvision.io.read_image(path)  # (C, H, W), dtype=torch.uint8, range [0,255]
            if img.shape[0] == 1:
                img = img.expand(3, -1, -1)
            frames.append(img.permute(1, 2, 0))  # (H, W, C)

        video_tensor = torch.stack(frames, dim=0)  # (T, H, W, 3)
        torchvision.io.write_video(video_path, video_tensor, fps=24)
        rich.print(f"[bold green]✅ Test video saved at[/bold green] [cyan]{video_path}[/cyan] with {len(frames)} frames")

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(params=self.parameters(), lr=self.cfg.lrate, eps=1e-15)
        scheduler = {
            "scheduler": torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, mode='min', factor=0.9, patience=10, min_lr=1e-5),
            "monitor": "loss",  # what to monitor for the scheduler
            "interval": "epoch",  # when to step the scheduler
            "frequency": 1,
        }
        return {"optimizer": optimizer, "lr_scheduler": scheduler}

    @staticmethod
    def generate_map(result_maps, pixels, pixels_highlighted, pixels_mask):
        rgb_map_column = torch.cat([rgb for rgb in result_maps['rgb_map']], dim=1)
        acc_map_column = torch.cat([rgb for rgb in result_maps['acc_map']], dim=1)
        depth_map_column = torch.cat([rgb for rgb in result_maps['depth_map']], dim=1)

        rgb_map_static_column = torch.cat([rgb for rgb in result_maps['rgb_map_static_independent']], dim=1) if 'rgb_map_static_independent' in result_maps else None
        acc_map_static_column = torch.cat([rgb for rgb in result_maps['acc_map_static_independent']], dim=1) if 'acc_map_static_independent' in result_maps else None
        depth_map_static_column = torch.cat([rgb for rgb in result_maps['depth_map_static_independent']], dim=1) if 'depth_map_static_independent' in result_maps else None

        rgb_map_dynamic_column = torch.cat([rgb for rgb in result_maps['rgb_map_dynamic_independent']], dim=1) if 'rgb_map_dynamic_independent' in result_maps else None
        acc_map_dynamic_column = torch.cat([rgb for rgb in result_maps['acc_map_dynamic_independent']], dim=1) if 'acc_map_dynamic_independent' in result_maps else None
        depth_map_dynamic_column = torch.cat([rgb for rgb in result_maps['depth_map_dynamic_independent']], dim=1) if 'depth_map_dynamic_independent' in result_maps else None

        pixels_column = torch.cat([pixel for pixel in pixels[..., :3]], dim=1) if pixels is not None else torch.zeros_like(rgb_map_column)
        pixels_highlighted_column = torch.cat([pixel for pixel in pixels_highlighted[..., :3]], dim=1) if pixels_highlighted is not None else torch.zeros_like(rgb_map_column)
        pixels_mask_column = torch.cat([mask for mask in pixels_mask.unsqueeze(-1)], dim=1) if pixels_mask is not None else torch.zeros_like(rgb_map_column)

        if rgb_map_static_column is not None:
            rgb_map_column = torch.cat([rgb_map_static_column, rgb_map_column], dim=0)
        if acc_map_static_column is not None:
            acc_map_column = torch.cat([acc_map_static_column, acc_map_column], dim=0)
        if depth_map_static_column is not None:
            depth_map_column = torch.cat([depth_map_static_column, depth_map_column], dim=0)
        if rgb_map_dynamic_column is not None:
            rgb_map_column = torch.cat([rgb_map_column, rgb_map_dynamic_column], dim=0)
        if acc_map_dynamic_column is not None:
            acc_map_column = torch.cat([acc_map_column, acc_map_dynamic_column], dim=0)
        if depth_map_dynamic_column is not None:
            depth_map_column = torch.cat([depth_map_column, depth_map_dynamic_column], dim=0)

        N = rgb_map_column.shape[0] // pixels_column.shape[0]
        if N > 1:
            left_side = torch.cat([pixels_highlighted_column, pixels_column, pixels_mask_column.expand_as(pixels_column)], dim=0)
        else:
            left_side = torch.cat([pixels_column, pixels_highlighted_column], dim=1)
        right_side = torch.cat([rgb_map_column, acc_map_column.expand_as(rgb_map_column), depth_map_column.expand_as(rgb_map_column)], dim=1)

        final_map = torch.cat([left_side, right_side], dim=1)
        return final_map

    def resimulate_explicit_fields(self, output_dir, hython_path, total_frames=120):
        aabb_std = self.renderer.aabb_std.cpu()  # [low_x, low_y, low_z, high_x, high_y, high_z]
        bound_std = self.renderer.bound_std.cpu()  # [low_x, low_y, low_z, high_x, high_y, high_z]

        weightx = (aabb_std[3] - aabb_std[0]).item() / 2
        weighty = (aabb_std[4] - aabb_std[1]).item() / 2
        weightz = (aabb_std[5] - aabb_std[2]).item() / 2

        min_bound = bound_std[:3]  # [low_x, low_y, low_z]
        max_bound = bound_std[3:]  # [high_x, high_y, high_z]

        base_res = 128
        resx, resy, resz = int(base_res * weightx), int(base_res * weighty), int(base_res * weightz)
        coords_x = torch.linspace(0, 1, steps=resx + 1)
        centers_x = (coords_x[:-1] + coords_x[1:]) / 2
        coords_y = torch.linspace(0, 1, steps=resy + 1)
        centers_y = (coords_y[:-1] + coords_y[1:]) / 2
        coords_z = torch.linspace(0, 1, steps=resz + 1)
        centers_z = (coords_z[:-1] + coords_z[1:]) / 2
        x, y, z = torch.meshgrid(centers_x, centers_y, centers_z, indexing='ij')
        xyz = torch.stack([x, y, z], dim=-1).view(-1, 3)
        mask_inside = ((xyz >= min_bound) & (xyz <= max_bound)).all(dim=-1)
        source_height = 0.15
        gt_mask = xyz[..., 1] < source_height
        xyz = xyz.to(torch.float16 if self.cfg.dataset.fp16 else torch.float32).to(self.device)
        mask_inside = mask_inside.to(self.device)
        gt_mask = gt_mask.reshape(resx, resy, resz).to(self.device)



        output_dir = os.path.join(output_dir, 'resimulate_fields')
        os.makedirs(output_dir, exist_ok=True)
        with torch.no_grad():
            with torch.amp.autocast('cuda', enabled=True):
                self.eval()
                sigmas_resimulated = []
                with rich.progress.Progress() as progress:
                    task = progress.add_task("[green]Resimulating...", total=total_frames)
                    dt = 1.0 / float(total_frames - 1)

                    self.model.set_time(time=torch.tensor([0], device=xyz.device, dtype=xyz.dtype))
                    den_sim_t_1 = self.sample_sigma(xyz=xyz, mask=mask_inside).reshape(resx, resy, resz, 1)
                    for frame_idx in range(total_frames):
                        if frame_idx > 0:
                            self.model.set_time(time=torch.tensor([frame_idx - 1 / total_frames], device=xyz.device, dtype=xyz.dtype))
                            vel_sim_t_1 = self.sample_velocity(xyz=xyz, mask=mask_inside).reshape(resx, resy, resz, 3)
                            self.model.set_time(time=torch.tensor([frame_idx / total_frames], device=xyz.device, dtype=xyz.dtype))
                            den_vis_t = self.sample_sigma(xyz=xyz, mask=mask_inside).reshape(resx, resy, resz, 1)
                            den_sim_t = advect_density(den=den_sim_t_1, vel=vel_sim_t_1, coord_3d_sim=xyz.reshape(resx, resy, resz, 3), dt=dt, gt=den_vis_t, gt_mask=gt_mask)
                        else:
                            den_sim_t = den_sim_t_1
                        sigmas_resimulated.append(den_sim_t.detach().cpu().clone())
                        den_sim_t_1 = den_sim_t
                        progress.update(task, advance=1)
                if hython_path is not None:
                    cpu_cores = max(os.cpu_count() - 3, 4)
                    rich.print(f"[bold green][Using multiprocessing with {cpu_cores} cores][/bold green], exporting resimulated fields to Houdini...")
                    with multiprocessing.Pool(processes=cpu_cores) as pool:
                        tasks = [
                            (hython_path, sigma, aabb_std, output_dir, frame_idx)
                            for frame_idx, sigma in enumerate(sigmas_resimulated)
                        ]
                        pool.starmap(self.export_density_houdini, tasks)
                else:
                    for frame_idx, sigma in enumerate(sigmas_resimulated):
                        torch.save(sigma, os.path.join(output_dir, f'sigma_{frame_idx:03d}.pt'))
                        rich.print(f"[bold green]✅ Simulated density field saved at[/bold green] [cyan]{os.path.join(output_dir, f'sigma_{frame_idx:03d}.pt')}[/cyan] with shape [magenta]{sigma.shape}[/magenta]")

        if hython_path is not None:
            gplay_exe = os.path.join(os.path.dirname(hython_path), 'gplay.exe')
            rich.print(f"[bold green]Opening gplay[/bold green]")
            subprocess.run([gplay_exe], check=True, cwd=f"{output_dir}")

    def export_explicit_fields(self, output_dir, hython_path, total_frames=120):
        aabb_std = self.renderer.aabb_std.cpu()  # [low_x, low_y, low_z, high_x, high_y, high_z]
        bound_std = self.renderer.bound_std.cpu()  # [low_x, low_y, low_z, high_x, high_y, high_z]

        weightx = (aabb_std[3] - aabb_std[0]).item() / 2
        weighty = (aabb_std[4] - aabb_std[1]).item() / 2
        weightz = (aabb_std[5] - aabb_std[2]).item() / 2

        min_bound = bound_std[:3]  # [low_x, low_y, low_z]
        max_bound = bound_std[3:]  # [high_x, high_y, high_z]

        base_res = 64
        resx, resy, resz = int(base_res * weightx), int(base_res * weighty), int(base_res * weightz)
        coords_x = torch.linspace(0, 1, steps=resx + 1)
        centers_x = (coords_x[:-1] + coords_x[1:]) / 2
        coords_y = torch.linspace(0, 1, steps=resy + 1)
        centers_y = (coords_y[:-1] + coords_y[1:]) / 2
        coords_z = torch.linspace(0, 1, steps=resz + 1)
        centers_z = (coords_z[:-1] + coords_z[1:]) / 2
        x, y, z = torch.meshgrid(centers_x, centers_y, centers_z, indexing='ij')
        xyz = torch.stack([x, y, z], dim=-1).view(-1, 3)
        mask_inside = ((xyz >= min_bound) & (xyz <= max_bound)).all(dim=-1)
        xyz = xyz.to(torch.float16 if self.cfg.dataset.fp16 else torch.float32).to(self.device)

        output_dir = os.path.join(output_dir, 'explicit_fields')
        os.makedirs(output_dir, exist_ok=True)
        with torch.no_grad():
            with torch.amp.autocast('cuda', enabled=True):
                self.eval()
                sigmas = []
                velocities = []
                with rich.progress.Progress() as progress:
                    task = progress.add_task("[green]Generating fields...", total=total_frames)
                    for frame_idx in range(total_frames):
                        self.model.set_time(time=torch.tensor([frame_idx / total_frames], device=xyz.device, dtype=xyz.dtype))
                        sigma = self.sample_sigma(xyz=xyz, mask=mask_inside).reshape(resx, resy, resz, 1)
                        velocity = self.sample_velocity(xyz=xyz, mask=mask_inside).reshape(resx, resy, resz, 3)
                        sigmas.append(sigma.cpu())
                        velocities.append(velocity.cpu())
                        progress.update(task, advance=1)
                if hython_path is not None:
                    cpu_cores = max(os.cpu_count() - 3, 4)
                    rich.print(f"[bold green][Using multiprocessing with {cpu_cores} cores][/bold green], exporting fields to Houdini...")
                    with multiprocessing.Pool(processes=cpu_cores) as pool:
                        tasks = [
                            (hython_path, sigma, aabb_std, output_dir, frame_idx)
                            for frame_idx, sigma in enumerate(sigmas)
                        ]
                        pool.starmap(self.export_density_houdini, tasks)
                    with multiprocessing.Pool(processes=cpu_cores) as pool:
                        tasks = [
                            (hython_path, velocity, aabb_std, output_dir, frame_idx)
                            for frame_idx, velocity in enumerate(velocities)
                        ]
                        pool.starmap(self.export_velocity_houdini, tasks)
                else:
                    for frame_idx, (sigma, velocity) in enumerate(zip(sigmas, velocities)):
                        torch.save(sigma, os.path.join(output_dir, f'sigma_{frame_idx:03d}.pt'))
                        rich.print(f"[bold green]✅ Density field saved at[/bold green] [cyan]{os.path.join(output_dir, f'sigma_{frame_idx:03d}.pt')}[/cyan] with shape [magenta]{sigma.shape}[/magenta]")
                        torch.save(velocity, os.path.join(output_dir, f'velocity_{frame_idx:03d}.pt'))
                        rich.print(f"[bold green]✅ Velocity field saved at[/bold green] [cyan]{os.path.join(output_dir, f'velocity_{frame_idx:03d}.pt')}[/cyan] with shape [magenta]{velocity.shape}[/magenta]")

        if hython_path is not None:
            gplay_exe = os.path.join(os.path.dirname(hython_path), 'gplay.exe')
            rich.print(f"[bold green]Opening gplay[/bold green]")
            subprocess.run([gplay_exe], check=True, cwd=f"{output_dir}")

    def sample_sigma(self, xyz, mask, chunk: int = 1024 * 8):
        sigma = []
        for start_idx in range(0, xyz.shape[0], chunk):
            ret = self.model.sigma(xyz[start_idx:start_idx + chunk])
            _sigma = ret[0]
            sigma.append(_sigma.detach())
        sigma = torch.cat(sigma, dim=0)
        if mask is not None:
            sigma[~mask] = 0.0
        return sigma

    def sample_velocity(self, xyz, mask, chunk: int = 1024 * 8):
        velocity = []
        for start_idx in range(0, xyz.shape[0], chunk):
            ret = self.model.vel(xyz[start_idx:start_idx + chunk])
            _velocity = ret
            velocity.append(_velocity.detach())
        velocity = torch.cat(velocity, dim=0)
        if mask is not None:
            velocity[~mask] = 0.0
        return velocity

    @staticmethod
    def export_density_houdini(hython_path, sigma, aabb_std, output_dir, frame_idx):
        low_x, low_y, low_z, high_x, high_y, high_z = aabb_std[0].item(), aabb_std[1].item(), aabb_std[2].item(), aabb_std[3].item(), aabb_std[4].item(), aabb_std[5].item()
        with tempfile.NamedTemporaryFile(suffix=".npy", delete=False) as f:
            temp_file = f.name
        np.save(temp_file, sigma.numpy())
        out_file = os.path.join(f"{output_dir}", f"sigma_{frame_idx:03d}.bgeo.sc")
        houdini_code = f'''
                        import hou
                        import numpy as np
                        import os
                        nparray = np.load(r"{temp_file}")
                        resx, resy, resz = nparray.shape[0:3]
                        geo = hou.Geometry()
                        name_attrib = geo.addAttrib(hou.attribType.Prim, "name", "default")
                        vol = geo.createVolume(resx, resy, resz, hou.BoundingBox(float({low_x}), float({low_y}), float({low_z}), float({high_x}), float({high_y}), float({high_z})))
                        vol.setAttribValue(name_attrib, "density")
                        vol.setAllVoxels(nparray.flatten().tolist())
                        geo.saveToFile(r"{out_file}")
                        '''
        try:
            EPINFTrainer.execute_template_script(hython_path=hython_path, script_content=houdini_code)
            rich.print(f"[bold green]✅ Exported[/bold green] [cyan]{out_file}[/cyan] with shape [magenta]{sigma.shape}[/magenta]")
        except subprocess.CalledProcessError as e:
            raise RuntimeError(f"Failed to execute Houdini script: {e}") from e
        finally:
            os.remove(temp_file)

    @staticmethod
    def export_velocity_houdini(hython_path, velocity, aabb_std, output_dir, frame_idx):
        low_x, low_y, low_z, high_x, high_y, high_z = aabb_std[0].item(), aabb_std[1].item(), aabb_std[2].item(), aabb_std[3].item(), aabb_std[4].item(), aabb_std[5].item()
        with tempfile.NamedTemporaryFile(suffix=".npy", delete=False) as f:
            temp_file = f.name
        np.save(temp_file, velocity.numpy())
        out_file = os.path.join(f"{output_dir}", f"velocity_{frame_idx:03d}.bgeo.sc")
        houdini_code = f'''
                        import hou
                        import numpy as np
                        import os
                        nparray = np.load(r"{temp_file}")
                        resx, resy, resz = nparray.shape[0:3]
                        geo = hou.Geometry()
                        name_attrib = geo.addAttrib(hou.attribType.Prim, "name", "default")
                        
                        for i, name in enumerate(['vel.x', 'vel.y', 'vel.z']):
                            vol = geo.createVolume(resx, resy, resz, hou.BoundingBox(float({low_x}), float({low_y}), float({low_z}), float({high_x}), float({high_y}), float({high_z})))
                            vol.setAttribValue(name_attrib, name)
                            vol.setAllVoxels(nparray[..., i].flatten().tolist())
                        geo.saveToFile(r"{out_file}")
                        '''
        try:
            EPINFTrainer.execute_template_script(hython_path=hython_path, script_content=houdini_code)
            rich.print(f"[bold green]✅ Exported[/bold green] [cyan]{out_file}[/cyan] with shape [magenta]{velocity.shape}[/magenta]")
        except subprocess.CalledProcessError as e:
            raise RuntimeError(f"Failed to execute Houdini script: {e}") from e
        finally:
            os.remove(temp_file)

    @staticmethod
    def execute_template_script(hython_path, script_content):
        with tempfile.NamedTemporaryFile(suffix=".py", delete=False, mode="w", encoding="utf-8") as f:
            temp_script_path = f.name
            f.write(textwrap.dedent(script_content))
        try:
            subprocess.run([hython_path, temp_script_path], check=True)
        except subprocess.CalledProcessError as e:
            raise RuntimeError(f"Failed to execute Houdini script: {e}") from e
        finally:
            os.remove(temp_script_path)



# -*- coding: utf-8 -*-
"""
Author: Xiang Yangcheng [https://github.com/Xayah-Hina]
GitHub: https://github.com/IME-lab-Hokudai/EPINF-NeuFlow
Last Modification Date: 2025-07-10
Description: The pytorch implementation of the paper "Efficient Physics Informed Neural Reconstruction of Diverse Fluid Dynamics from Sparse Observations".
License: Mozilla Public License 2.0 (MPL-2.0)
Copyright (c) 2025 Xiang Yangcheng, IME Lab, Hokkaido University, Japan.
"""
import os
import typing

import lightning
import torch


class PIDataset(torch.utils.data.Dataset):
    """
    PRIVATE CLASS
    - This class and its subclasses are only used inside datasets package.
    """

    @property
    def images(self):
        raise NotImplementedError("This method should be implemented in subclasses.")

    @property
    def images_highlighted(self):
        raise NotImplementedError("This method should be implemented in subclasses.")

    @property
    def images_masks(self):
        raise NotImplementedError("This method should be implemented in subclasses.")

    @property
    def poses(self):
        raise NotImplementedError("This method should be implemented in subclasses.")

    @property
    def focal(self):
        raise NotImplementedError("This method should be implemented in subclasses.")

    @property
    def width(self):
        raise NotImplementedError("This method should be implemented in subclasses.")

    @property
    def height(self):
        raise NotImplementedError("This method should be implemented in subclasses.")

    @property
    def near(self):
        raise NotImplementedError("This method should be implemented in subclasses.")

    @property
    def far(self):
        raise NotImplementedError("This method should be implemented in subclasses.")

    @property
    def num_poses(self):
        raise NotImplementedError("This method should be implemented in subclasses.")

    @property
    def num_times(self):
        raise NotImplementedError("This method should be implemented in subclasses.")

    def transform(self, translate, scale):
        raise NotImplementedError("This method should be implemented in subclasses.")


class PIDatasetAPI(lightning.LightningDataModule):
    """
    PUBLIC API:
        dataset_path: str
        aabb_std: list[min_x, min_y, min_z, max_x, max_y, max_z]
        bound_std: list[min_x, min_y, min_z, max_x, max_y, max_z]
        background_color: list[r, g, b]
        complex_background: bool
        num_train_poses: int
    """

    def __init__(self, cfg):
        super().__init__()
        self.cfg = cfg

        self.dataset_train: PIDataset | None = None
        self.dataset_val: PIDataset | None = None
        self.dataset_test: PIDataset | None = None

    def prepare_data(self):
        raise NotImplementedError("This method should be implemented in subclasses.")

    def setup(self, stage: str):
        if stage == 'fit':
            self.dataset_train = self.dataset('train')
            self.dataset_val = self.dataset('val')
        elif stage == 'validate':
            self.dataset_val = self.dataset('val')
        elif stage == 'test' or stage == 'predict':
            self.dataset_test = self.dataset('test')
        else:
            raise NotImplementedError('Unsupported stage: {}'.format(stage))

    def dataset(self, dataset_type: typing.Literal['train', 'val', 'test']) -> PIDataset:
        raise NotImplementedError("This method should be implemented in subclasses.")

    @staticmethod
    def collate(batch: list):
        raise NotImplementedError("This method should be implemented in subclasses.")

    def train_dataloader(self):
        return torch.utils.data.DataLoader(
            self.dataset_train,
            batch_size=1,
            shuffle=True,
            num_workers=min(os.cpu_count() - 1, 4),
            persistent_workers=True,
            collate_fn=self.collate,
        )

    def val_dataloader(self):
        return torch.utils.data.DataLoader(
            self.dataset_val,
            batch_size=1,
            shuffle=False,
            num_workers=min(os.cpu_count() - 1, 4),
            persistent_workers=True,
            collate_fn=self.collate,
        )

    def test_dataloader(self):
        return torch.utils.data.DataLoader(
            self.dataset_test,
            batch_size=1,
            shuffle=False,
            num_workers=min(os.cpu_count() - 1, 4),
            persistent_workers=True,
            collate_fn=self.collate,
        )

    def predict_dataloader(self):
        return self.test_dataloader()

    @property
    def pre_translate(self):
        raise NotImplementedError("This method should be implemented in subclasses.")

    @property
    def pre_scale(self):
        raise NotImplementedError("This method should be implemented in subclasses.")

    @property
    def aabb_std(self):
        raise NotImplementedError("This method should be implemented in subclasses.")

    @property
    def bound_std(self):
        raise NotImplementedError("This method should be implemented in subclasses.")

    @property
    def background_color(self):
        raise NotImplementedError("This method should be implemented in subclasses.")

    @property
    def complex_background(self):
        raise NotImplementedError("This method should be implemented in subclasses.")

    @property
    def num_train_poses(self):
        assert self.dataset_train is not None, "Dataset for training is not set up."
        return self.dataset_train.num_poses

    @property
    def width(self):
        assert self.dataset_train is not None, "Dataset for training is not set up."
        return self.dataset_train.width

    @property
    def height(self):
        assert self.dataset_train is not None, "Dataset for training is not set up."
        return self.dataset_train.height

    @property
    def near_std(self):
        assert self.dataset_train is not None, "Dataset for training is not set up."
        return self.dataset_train.near

    @property
    def far_std(self):
        assert self.dataset_train is not None, "Dataset for training is not set up."
        return self.dataset_train.far
